from __future__ import annotations

import typing
from copy import copy, deepcopy
from typing import Iterable, List, overload

import pandas as pd

from .exceptions import PlotnineError
from .mapping.aes import NO_GROUP, SCALED_AESTHETICS, aes
from .mapping.evaluation import evaluate, stage
from .utils import array_kind, check_required_aesthetics, ninteraction

if typing.TYPE_CHECKING:
    from typing import Any, Optional, Sequence, SupportsIndex

    from plotnine.typing import (
        Coord,
        DataFrameConvertible,
        DataLike,
        EvalEnvironment,
        Geom,
        Ggplot,
        Layer,
        LayerDataLike,
        Layout,
        Position,
        Scales,
        Stat,
    )


class layer:
    """
    Layer

    When a ``geom`` or ``stat`` is added to a
    :class:`~plotnine.ggplot` object, it creates a single layer.
    This class is a representation of that layer.

    Parameters
    ----------
    geom : geom, optional
        geom to used to draw this layer.
    stat : stat, optional
        stat used for the statistical transformation of
        data in this layer
    mapping : aes, optional
        Aesthetic mappings.
    data : dataframe, optional
        Data plotted in this layer. If ``None``, the data from
        the :class:`~plotnine.ggplot` object will be used.
    position : position, optional
        Position object to adjust the geometries in this layer.
    inherit_aes : bool, optional
        If ``True`` inherit from the aesthetic mappings of
        the :class:`~plotnine.ggplot` object. Default ``True``.
    show_legend : bool or None, optional
        Whether to make up and show a legend for the mappings
        of this layer. If ``None`` then an automatic/good choice
        is made. Default is ``None``.
    raster : bool, optional (default: False)
        If ``True``, draw onto this layer a raster (bitmap) object
        even if the final image format is vector.

    Notes
    -----
    There is no benefit to manually creating a layer. You should
    always use a ``geom`` or ``stat``.
    """

    # Data for this layer
    data: pd.DataFrame

    def __init__(
        self,
        geom: Geom,
        stat: Stat,
        *,
        mapping: aes,
        data: Optional[LayerDataLike],
        position: Position,
        inherit_aes: bool = True,
        show_legend: bool | None = None,
        raster: bool = False,
    ):
        self.geom = geom
        self.stat = stat
        self._data = data
        self.mapping = mapping
        self.position = position
        self.inherit_aes = inherit_aes
        self.show_legend = show_legend
        self.raster = raster
        self.zorder = 0

    @staticmethod
    def from_geom(geom: Geom) -> Layer:
        """
        Create a layer given a :class:`geom`

        Parameters
        ----------
        geom : geom
            `geom` from which a layer will be created

        Returns
        -------
        out : layer
            Layer that represents the specific `geom`.
        """
        kwargs = geom._kwargs
        lkwargs = {
            "geom": geom,
            "mapping": geom.mapping,
            "data": geom.data,
            "stat": geom._stat,
            "position": geom._position,
        }

        layer_params = ("inherit_aes", "show_legend", "raster")
        for param in layer_params:
            if param in kwargs:
                lkwargs[param] = kwargs[param]
            elif param in geom.DEFAULT_PARAMS:
                lkwargs[param] = geom.DEFAULT_PARAMS[param]
        return layer(**lkwargs)

    def __radd__(self, gg: Ggplot) -> Ggplot:
        """
        Add layer to ggplot object
        """
        try:
            gg.layers.append(self)
        except AttributeError:
            msg = "Cannot add layer to object of type {!r}".format
            raise PlotnineError(msg(type(gg)))
        return gg

    def __deepcopy__(self, memo: dict[Any, Any]) -> layer:
        """
        Deep copy without copying the self.data dataframe
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        for key, item in old.items():
            if key == "data":
                new[key] = old[key]
            else:
                new[key] = deepcopy(old[key], memo)

        return result

    def setup(self, plot: Ggplot):
        """
        Prepare layer for the plot building

        Give the layer access to the data, mapping and environment
        """
        self._make_layer_data(plot.data)
        self._make_layer_mapping(plot.mapping)
        self._make_layer_environments(plot.environment)

    def _make_layer_data(self, plot_data: DataLike | None):
        """
        Generate data to be used by this layer

        Parameters
        ----------
        plot_data : dataframe
            ggplot object data
        """
        if plot_data is None:
            data = pd.DataFrame()
        elif hasattr(plot_data, "to_pandas"):
            data = typing.cast("DataFrameConvertible", plot_data).to_pandas()
        else:
            data = typing.cast("pd.DataFrame", plot_data)

        # Each layer that does not have data gets a copy of
        # of the ggplot.data. If it has data it is replaced
        # by copy so that we do not alter the users data
        if self._data is None:
            try:
                self.data = copy(data)
            except AttributeError:
                _geom_name = self.geom.__class__.__name__
                _data_name = data.__class__.__name__
                raise PlotnineError(
                    f"{_geom_name} layer expects a dataframe, "
                    f"but it got {_data_name} instead."
                )
        elif callable(self._data):
            self.data = self._data(data)
            if not isinstance(self.data, pd.DataFrame):
                raise PlotnineError(
                    "Data function must return a Pandas dataframe"
                )
        else:
            # Recognise polars dataframes
            if hasattr(self._data, "to_pandas"):
                self.data = typing.cast(
                    "DataFrameConvertible", self._data
                ).to_pandas()
            elif isinstance(self._data, pd.DataFrame):
                self.data = self._data.copy()
            else:
                raise TypeError(f"Data has a bad type: {type(self.data)}")

    def _make_layer_mapping(self, plot_mapping: aes):
        """
        Create the aesthetic mappings to be used by this layer

        Parameters
        ----------
        plot_mapping : aes
            ggplot object mapping
        """
        if self.inherit_aes:
            self.mapping = self.mapping.inherit(plot_mapping)

        # aesthetics set as parameters override the same
        # aesthetics set as mappings, so we can ignore
        # those in the mapping
        for ae in self.geom.aes_params:
            if ae in self.mapping:
                del self.mapping[ae]

        # Set group as a mapping if set as a parameter
        if "group" in self.geom.aes_params:
            group = self.geom.aes_params["group"]
            # Double quote str so that it evaluates to itself
            if isinstance(group, str):
                group = f'"{group}"'
            self.mapping["group"] = stage(start=group)

    def _make_layer_environments(self, plot_environment: EvalEnvironment):
        """
        Create the aesthetic mappings to be used by this layer

        Parameters
        ----------
        plot_environment : ~patsy.Eval.EvalEnvironment
            Namespace in which to execute aesthetic expressions.
        """
        self.geom.environment = plot_environment
        self.stat.environment = plot_environment

    def compute_aesthetics(self, plot: Ggplot):
        """
        Return a dataframe where the columns match the aesthetic mappings

        Transformations like 'factor(cyl)' and other
        expression evaluation are  made in here
        """
        evaled = evaluate(self.mapping._starting, self.data, plot.environment)
        evaled_aes = aes(**{str(col): col for col in evaled})
        plot.scales.add_defaults(evaled, evaled_aes)

        if len(self.data) == 0 and len(evaled) > 0:
            # No data, and vectors suppled to aesthetics
            evaled["PANEL"] = 1
        else:
            evaled["PANEL"] = self.data["PANEL"]

        data = add_group(evaled)
        self.data = data.sort_values("PANEL", kind="mergesort")

    def compute_statistic(self, layout: Layout):
        """
        Compute & return statistics for this layer
        """
        data = self.data
        if not len(data):
            return

        params = self.stat.setup_params(data)
        data = self.stat.use_defaults(data)
        data = self.stat.setup_data(data)
        data = self.stat.compute_layer(data, params, layout)
        self.data = data

    def map_statistic(self, plot: Ggplot):
        """
        Mapping aesthetics to computed statistics
        """
        data = self.data
        if not len(data):
            return

        # Mixin default stat aesthetic mappings
        aesthetics = self.mapping.inherit(self.stat.DEFAULT_AES)
        stat_data = evaluate(aesthetics._calculated, data, plot.environment)

        if not len(stat_data):
            return

        # (see stat_spoke for one exception)
        if self.stat.retransform:
            stat_data = plot.scales.transform_df(stat_data)

        # When there are duplicate columns, we use the computed
        # ones in stat_data
        columns = data.columns.difference(stat_data.columns)
        self.data = pd.concat([data[columns], stat_data], axis=1)

        # Add any new scales, if needed
        new = {ae: ae for ae in stat_data.columns}
        plot.scales.add_defaults(self.data, new)

    def setup_data(self):
        """
        Prepare/modify data for plotting
        """
        data = self.data
        if len(data) == 0:
            return

        data = self.geom.setup_data(data)

        check_required_aesthetics(
            self.geom.REQUIRED_AES,
            set(data.columns) | set(self.geom.aes_params),
            self.geom.__class__.__name__,
        )

        self.data = data

    def compute_position(self, layout: Layout):
        """
        Compute the position of each geometric object

        This is in concert with the other objects in the panel depending
        on the position class of the geom
        """
        if len(self.data) == 0:
            return

        params = self.position.setup_params(self.data)
        data = self.position.setup_data(self.data, params)
        data = self.position.compute_layer(data, params, layout)
        self.data = data

    def draw(self, layout: Layout, coord: Coord):
        """
        Draw geom

        Parameters
        ----------
        layout : Layout
            Layout object created when the plot is getting
            built
        coord : coord
            Type of coordinate axes
        """
        params = copy(self.geom.params)
        params.update(self.stat.params)
        params["zorder"] = self.zorder
        params["raster"] = self.raster
        self.data = self.geom.handle_na(self.data)
        # At this point each layer must have the data
        # that is created by the plot build process
        self.geom.draw_layer(self.data, layout, coord, **params)

    def use_defaults(
        self,
        data: pd.DataFrame | None = None,
        aes_modifiers: dict[str, Any] | None = None,
    ) -> pd.DataFrame:
        """
        Prepare/modify data for plotting

        Parameters
        ----------
        data : dataframe, optional
            Data
        aes_modifiers : dict
            Expression to evaluate and replace aesthetics in
            the data.
        """
        if data is None:
            data = self.data

        if aes_modifiers is None:
            aes_modifiers = self.mapping._scaled

        return self.geom.use_defaults(data, aes_modifiers)

    def finish_statistics(self):
        """
        Prepare/modify data for plotting
        """
        self.stat.finish_layer(self.data, self.stat.params)


class Layers(List[layer]):
    """
    List of layers

    During the plot building pipeline, many operations are
    applied at all layers in the plot. This class makes those
    tasks easier.
    """

    @overload
    def __radd__(self, other: Iterable[layer]) -> Layers:
        ...

    @overload
    def __radd__(self, other: Ggplot) -> Ggplot:
        ...

    def __radd__(self, other: Iterable[layer] | Ggplot) -> Layers | Ggplot:
        """
        Add layers to ggplot object
        """
        # Add layers to ggplot object
        from .ggplot import ggplot

        if isinstance(other, ggplot):
            for obj in self:
                other += obj
        else:
            raise PlotnineError(
                f"Cannot add Layers to object of type {type(other)}"
            )
        return other

    @overload
    def __getitem__(self, key: SupportsIndex) -> layer:
        ...

    @overload
    def __getitem__(self, key: slice) -> Layers:
        ...

    def __getitem__(self, key: SupportsIndex | slice) -> layer | Layers:
        result = super().__getitem__(key)
        if isinstance(result, Iterable):
            result = Layers(result)
        return result

    @property
    def data(self) -> list[pd.DataFrame]:
        return [l.data for l in self]

    def setup(self, plot: Ggplot):
        for l in self:
            l.setup(plot)

    def setup_data(self):
        for l in self:
            l.setup_data()

    def draw(self, layout: Layout, coord: Coord):
        # If zorder is 0, it is left to MPL
        for i, l in enumerate(self, start=1):
            l.zorder = i
            l.draw(layout, coord)

    def compute_aesthetics(self, plot: Ggplot):
        for l in self:
            l.compute_aesthetics(plot)

    def compute_statistic(self, layout: Layout):
        for l in self:
            l.compute_statistic(layout)

    def map_statistic(self, plot: Ggplot):
        for l in self:
            l.map_statistic(plot)

    def compute_position(self, layout: Layout):
        for l in self:
            l.compute_position(layout)

    def use_defaults(
        self,
        data: pd.DataFrame | None = None,
        aes_modifiers: dict[str, Any] | None = None,
    ):
        for l in self:
            l.use_defaults(data, aes_modifiers)

    def transform(self, scales: Scales):
        for l in self:
            l.data = scales.transform_df(l.data)

    def train(self, scales: Scales):
        for l in self:
            scales.train_df(l.data)

    def map(self, scales: Scales):
        for l in self:
            l.data = scales.map_df(l.data)

    def finish_statistics(self):
        for l in self:
            l.finish_statistics()

    def update_labels(self, plot: Ggplot):
        for l in self:
            plot._update_labels(l)


def add_group(data: pd.DataFrame) -> pd.DataFrame:
    """
    Add group to the dataframe

    The group depends on the interaction of the discrete
    aesthetic columns in the dataframe.
    """
    if len(data) == 0:
        return data

    if "group" not in data:
        ignore = data.columns.difference(list(SCALED_AESTHETICS))
        disc = discrete_columns(data, ignore=ignore)
        if disc:
            data["group"] = ninteraction(data[disc], drop=True)
        else:
            data["group"] = NO_GROUP
    else:
        data["group"] = ninteraction(data[["group"]], drop=True)

    return data


def discrete_columns(
    df: pd.DataFrame, ignore: Sequence[str] | pd.Index
) -> list[str]:
    """
    Return a list of the discrete columns in the dataframe

    Parameters
    ----------
    df : dataframe
        Data
    ignore : list[str]
        A list|set|tuple with the names of the columns to skip.
    """
    lst = []
    for col in df:
        if array_kind.discrete(df[col]) and (col not in ignore):
            # Some columns are represented as object dtype
            # but may have compound structures as values.
            try:
                hash(df[col].iloc[0])
            except TypeError:
                continue
            lst.append(str(col))
    return lst



import pandas as pd

from plotnine import aes, geom_col, ggplot

df = pd.DataFrame(
    {
        "x": pd.Categorical(["b", "d", "c", "a"], ordered=True),
        "y": [1, 2, 3, 4],
    }
)


def test_reorder():
    p = (
        ggplot(df, aes("reorder(x, y)", "y", fill="reorder(x, y)"))
        + geom_col()
    )
    assert p == "reorder"


def test_reorder_index():
    # The dataframe is created with ordering according to the y
    # variable. So the x index should be ordered acc. to y too
    p = ggplot(df, aes("reorder(x, x.index)", "y")) + geom_col()
    assert p == "reorder_index"


def test_labels_series():
    p = ggplot(df, aes(x=df.x, y=df.y)) + geom_col()
    assert p.labels.x == "x"
    assert p.labels.y == "y"


def test_labels_lists():
    p = ggplot(df, aes(x=[1, 2, 3], y=[1, 2, 3])) + geom_col()
    assert p.labels.x is None
    assert p.labels.y is None



import numpy as np
import pandas as pd
import pytest
from mizani.transforms import log_format, log_trans

from plotnine import (
    aes,
    annotation_logticks,
    coord_flip,
    element_line,
    facet_wrap,
    geom_point,
    ggplot,
    scale_x_continuous,
    scale_x_log10,
    scale_y_log10,
    theme,
)
from plotnine.exceptions import PlotnineWarning

df = pd.DataFrame({"x": 10 ** np.arange(4)})


def test_annotation_logticks():
    # The grid should align with the logticks
    p = (
        ggplot(df, aes("x", "x"))
        + annotation_logticks(sides="b", size=0.75)
        + geom_point()
        + scale_x_log10()
        + scale_y_log10()
        + theme(
            panel_grid_minor=element_line(color="green"),
            panel_grid_major=element_line(color="red"),
        )
    )

    assert p == "annotation_logticks"


def test_annotation_logticks_faceting():
    n = len(df)
    df2 = pd.DataFrame(
        {"x": np.hstack([df["x"], df["x"]]), "g": list("a" * n + "b" * n)}
    )
    p = (
        ggplot(df2)
        + annotation_logticks(sides="b", size=0.75)
        + geom_point(aes("x", "x"))
        + scale_x_log10()
        + scale_y_log10()
        + facet_wrap("g")
        + theme(
            panel_grid_minor=element_line(color="green"),
            panel_grid_major=element_line(color="red"),
        )
    )

    assert p == "annotation_logticks_faceting"


def test_annotation_logticks_coord_flip():
    p = (
        ggplot(df, aes("x", "x"))
        + annotation_logticks(sides="b", size=0.75)
        + geom_point()
        + scale_x_log10()
        + scale_y_log10()
        + coord_flip()
        + theme(
            panel_grid_minor=element_line(color="green"),
            panel_grid_major=element_line(color="red"),
        )
    )

    assert p == "annotation_logticks_coord_flip"


def test_annotation_logticks_coord_flip_discrete():
    df = pd.DataFrame({"x": 10.0 ** (np.arange(4) - 1)})
    df2 = df.assign(
        discrete=pd.Categorical(["A" + str(int(a)) for a in df["x"]])
    )
    # the range on the 'A' range is 0..1,
    # but 0.1..100 on the y
    # to verify we are using the correct range for the log ticks
    df2 = df2.drop(df2.index[1:3])
    p = (
        ggplot(df2, aes("discrete", "x"))
        + annotation_logticks(sides="l", size=0.75)
        + geom_point()
        + scale_y_log10()
        + coord_flip()
        + theme(
            panel_grid_minor=element_line(color="green"),
            panel_grid_major=element_line(color="red"),
        )
    )

    assert p == "annotation_logticks_coord_flip_discrete"


def test_annotation_logticks_coord_flip_discrete_bottom():
    df = pd.DataFrame({"x": 10.0 ** (np.arange(4) - 1)})
    df2 = df.assign(
        discrete=pd.Categorical(["A" + str(int(a)) for a in df["x"]])
    )
    # the range on the 'A' range is 0..1,
    # but 0.1..100 on the y
    # to verify we are using the correct range for the log ticks
    df2 = df2.drop(df2.index[1:3])
    p = (
        ggplot(df2, aes("x", "discrete"))
        + annotation_logticks(sides="b", size=0.75)
        + geom_point()
        + scale_x_log10()
        + coord_flip()
        + theme(
            panel_grid_minor=element_line(color="green"),
            panel_grid_major=element_line(color="red"),
        )
    )

    assert p == "annotation_logticks_coord_flip_discrete_bottom"


def test_annotation_logticks_base_8():
    base = 8
    df = pd.DataFrame({"x": base ** np.arange(4)})
    # The grid should align with the logticks
    p = (
        ggplot(df, aes("x", "x"))
        + annotation_logticks(sides="b", size=0.75)
        + geom_point()
        + scale_x_continuous(
            trans=log_trans(base=base),
            labels=log_format(base=base, mathtex=True),
        )
        + theme(
            panel_grid_minor=element_line(color="green"),
            panel_grid_major=element_line(color="red"),
        )
    )

    assert p == "annotation_logticks_base_8"


def test_annotation_logticks_base_5():
    base = 5
    df = pd.DataFrame({"x": base ** np.arange(4)})
    # The grid should align with the logticks, and the logticks
    # should not have a midpoint
    p = (
        ggplot(df, aes("x", "x"))
        + annotation_logticks(sides="b", size=0.75)
        + geom_point()
        + scale_x_continuous(trans=log_trans(base=base))
        + theme(
            panel_grid_minor=element_line(color="green"),
            panel_grid_major=element_line(color="red"),
        )
    )

    assert p == "annotation_logticks_base_5"


def test_wrong_bases():
    # x axis not transformed
    p = (
        ggplot(df, aes("x", "x"))
        + annotation_logticks(sides="b", size=0.75, base=10)
        + geom_point()
    )

    with pytest.warns(PlotnineWarning):
        p.draw_test()

    # x axis not transform, but ticks requested for a different base
    p = (
        ggplot(df, aes("x", "x"))
        + annotation_logticks(sides="b", size=0.75, base=10)
        + scale_x_continuous(trans=log_trans(8))
        + geom_point()
    )

    with pytest.warns(PlotnineWarning):
        p.draw_test()

    # x axis is discrete
    df2 = df.assign(discrete=pd.Categorical([str(a) for a in df["x"]]))
    p = (
        ggplot(df2, aes("discrete", "x"))
        + annotation_logticks(sides="b", size=0.75, base=None)
        + geom_point()
    )

    with pytest.warns(PlotnineWarning):
        p.draw_test()

    # y axis is discrete
    df2 = df.assign(discrete=pd.Categorical([str(a) for a in df["x"]]))
    p = (
        ggplot(df2, aes("x", "discrete"))
        + annotation_logticks(sides="l", size=0.75, base=None)
        + geom_point()
    )

    with pytest.warns(PlotnineWarning):
        p.draw_test()

    # x axis is discrete + coord flip.
    df2 = df.assign(discrete=pd.Categorical([str(a) for a in df["x"]]))
    p = (
        ggplot(df2, aes("discrete", "x"))
        + annotation_logticks(sides="b", size=0.75, base=None)
        + geom_point()
        + coord_flip()
    )

    with pytest.warns(PlotnineWarning):
        p.draw_test()

    # y axis is discrete + coord_flip
    df2 = df.assign(discrete=pd.Categorical([str(a) for a in df["x"]]))
    p = (
        ggplot(df2, aes("x", "discrete"))
        + annotation_logticks(sides="l", size=0.75, base=None)
        + geom_point()
        + coord_flip()
    )

    with pytest.warns(PlotnineWarning):
        p.draw_test()



import numpy as np
import pandas as pd

from ..doctools import document
from .density import get_var_type, kde
from .stat import stat


@document
class stat_density_2d(stat):
    """
    Compute 2D kernel density estimation

    {usage}

    Parameters
    ----------
    {common_parameters}
    contour : bool
        Whether to create contours of the 2d density estimate.
        Default is True.
    n : int, optional(default: 64)
        Number of equally spaced points at which the density is to
        be estimated. For efficient computation, it should be a power
        of two.
    levels : int or array_like
        Contour levels. If an integer, it specifies the maximum number
        of levels, if array_like it is the levels themselves. Default
        is 5.
    package : str in ``['statsmodels', 'scipy', 'sklearn']``
        Package whose kernel density estimation to use. Default is
        statsmodels.
    kde_params : dict
        Keyword arguments to pass on to the kde class.

    See Also
    --------
    statsmodels.nonparametric.kde.KDEMultivariate
    scipy.stats.gaussian_kde
    sklearn.neighbors.KernelDensity
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'level'     # density level of a contour
        'density'   # Computed density at a point
        'piece'     # Numeric id of a contour in a given group

    `level` is only relevant when contours are computed. `density`
    is available only when no contours are computed. `piece` is
    largely irrelevant.
    """
    REQUIRED_AES = {"x"}
    DEFAULT_PARAMS = {
        "geom": "density_2d",
        "position": "identity",
        "na_rm": False,
        "contour": True,
        "package": "statsmodels",
        "kde_params": None,
        "n": 64,
        "levels": 5,
    }
    CREATES = {"y"}

    def setup_params(self, data):
        params = self.params.copy()
        if params["kde_params"] is None:
            params["kde_params"] = {}

        kde_params = params["kde_params"]
        if params["package"] == "statsmodels":
            params["package"] = "statsmodels-m"
            if "var_type" not in kde_params:
                kde_params["var_type"] = "{}{}".format(
                    get_var_type(data["x"]), get_var_type(data["y"])
                )

        return params

    @classmethod
    def compute_group(cls, data, scales, **params):
        package = params["package"]
        kde_params = params["kde_params"]

        group = data["group"].iloc[0]
        range_x = scales.x.dimension()
        range_y = scales.y.dimension()
        x = np.linspace(range_x[0], range_x[1], params["n"])
        y = np.linspace(range_y[0], range_y[1], params["n"])

        # The grid must have a "similar" shape (n, p) to the var_data
        X, Y = np.meshgrid(x, y)
        var_data = np.array([data["x"].values, data["y"].values]).T
        grid = np.array([X.flatten(), Y.flatten()]).T
        density = kde(var_data, grid, package, **kde_params)

        if params["contour"]:
            Z = density.reshape(len(x), len(y))
            data = contour_lines(X, Y, Z, params["levels"])
            # Each piece should have a distinct group
            groups = str(group) + "-00" + data["piece"].astype(str)
            data["group"] = groups
        else:
            data = pd.DataFrame(
                {
                    "x": X.flatten(),
                    "y": Y.flatten(),
                    "density": density.flatten(),
                    "group": group,
                    "level": 1,
                    "piece": 1,
                }
            )

        return data


def contour_lines(X, Y, Z, levels):
    """
    Calculate contour lines
    """
    from contourpy import contour_generator

    # Preparation of values and the creating of contours is
    # adapted from MPL with some adjustments.
    X = np.asarray(X, dtype=np.float64)
    Y = np.asarray(Y, dtype=np.float64)
    Z = np.asarray(Z, dtype=np.float64)
    zmin, zmax = Z.min(), Z.max()
    cgen = contour_generator(
        X, Y, Z, name="mpl2014", corner_mask=False, chunk_size=0
    )

    if isinstance(levels, int):
        from mizani.breaks import extended_breaks

        levels = extended_breaks(n=levels)((zmin, zmax))

    # The counter_generator gives us a list of vertices that
    # represent all the contour lines at that level. There
    # may be 0, 1 or more vertices at a level. Each one of
    # these we call a piece, and it represented as an nx2 array.
    #
    # We want x-y values that describe *all* the contour lines
    # in tidy format. Therefore each x-y vertex has a
    # corresponding level and piece id.
    segments = []
    piece_ids = []
    level_values = []
    start_pid = 1
    for level in levels:
        vertices, _ = cgen.create_contour(level)
        for pid, piece in enumerate(vertices, start=start_pid):
            n = len(piece)
            segments.append(piece)
            piece_ids.append(np.repeat(pid, n))
            level_values.append(np.repeat(level, n))
            start_pid = pid + 1

    # Collapse the info and make it fit for dataframe columns
    if segments:
        x, y = np.vstack(segments).T
        piece = np.hstack(piece_ids)
        level = np.hstack(level_values)
    else:
        x, y = [], []
        piece = []
        level = []

    data = pd.DataFrame(
        {
            "x": x,
            "y": y,
            "level": level,
            "piece": piece,
        }
    )
    return data



"""
Theme elements used to decorate the graph.
"""
from __future__ import annotations

import typing
from contextlib import suppress
from dataclasses import dataclass

if typing.TYPE_CHECKING:
    from typing import Any, Callable, Literal, Optional, Sequence

    from plotnine.typing import Theme, TupleFloat3, TupleFloat4


class element_base:
    """
    Base class for all theme elements
    """

    properties: dict[str, Any]  # dict of the properties

    def __init__(self):
        self.properties = {"visible": True}

    def __repr__(self) -> str:
        """
        Element representation
        """
        return f"{self.__class__.__name__}({self})"

    def __str__(self) -> str:
        """
        Element as string
        """
        d = self.properties.copy()
        del d["visible"]
        return f"{d}"

    def setup(self, theme: Theme):
        """
        Setup the theme_element before drawing
        """
        pass


class element_line(element_base):
    """
    theme element: line

    used for backgrounds and borders

    parameters
    ----------
    color : str | tuple
        line color
    colour : str | tuple
        alias of color
    linetype : str | tuple
        line style. if a string, it should be one of *solid*, *dashed*,
        *dashdot* or *dotted*. you can create interesting dashed patterns
        using tuples, see :meth:`matplotlib.lines.line2D.set_linestyle`.
    size : float
        line thickness
    kwargs : dict
        parameters recognised by
        :class:`matplotlib.lines.line2d`.
    """

    def __init__(
        self,
        *,
        color: Optional[str | TupleFloat3 | TupleFloat4] = None,
        size: Optional[float] = None,
        linetype: Optional[str | Sequence[int]] = None,
        lineend: Optional[Literal["butt", "projecting", "round"]] = None,
        colour: Optional[str | TupleFloat3 | TupleFloat4] = None,
        **kwargs: Any,
    ):
        super().__init__()
        self.properties.update(**kwargs)

        color = color if color else colour
        if color:
            self.properties["color"] = color
        if size:
            self.properties["linewidth"] = size
        if linetype:
            self.properties["linestyle"] = linetype

        if linetype in ("solid", "-") and lineend:
            self.properties["solid_capstyle"] = lineend
        elif linetype and lineend:
            self.properties["dash_capstyle"] = lineend


class element_rect(element_base):
    """
    Theme element: Rectangle

    Used for backgrounds and borders

    Parameters
    ----------
    fill : str | tuple
        Rectangle background color
    color : str | tuple
        Line color
    colour : str | tuple
        Alias of color
    size : float
        Line thickness
    kwargs : dict
        Parameters recognised by
        :class:`matplotlib.patches.Rectangle`. In some cases
        you can use the fancy parameters from
        :class:`matplotlib.patches.FancyBboxPatch`.
    """

    def __init__(
        self,
        fill: Optional[str | TupleFloat3 | TupleFloat4] = None,
        color: Optional[str | TupleFloat3 | TupleFloat4] = None,
        size: Optional[float] = None,
        linetype: Optional[str | Sequence[int]] = None,
        colour: Optional[str | TupleFloat3 | TupleFloat4] = None,
        **kwargs: Any,
    ):
        super().__init__()
        self.properties.update(**kwargs)

        color = color if color else colour
        if fill:
            self.properties["facecolor"] = fill
        if color:
            self.properties["edgecolor"] = color
        if size:
            self.properties["linewidth"] = size
        if linetype:
            self.properties["linestyle"] = linetype


class element_text(element_base):
    """
    Theme element: Text

    Parameters
    ----------
    family : str
        Font family. See :meth:`matplotlib.text.Text.set_family`
        for supported values.
    style : str in ``['normal', 'italic', 'oblique']``
        Font style
    color : str | tuple
        Text color
    weight : str
        Should be one of *normal*, *bold*, *heavy*, *light*,
        *ultrabold* or *ultralight*.
    size : float
        text size
    ha : str in ``['center', 'left', 'right']``
        Horizontal Alignment.
    va : str in ``['center' , 'top', 'bottom', 'baseline']``
        Vertical alignment.
    rotation : float
        Rotation angle in the range [0, 360]
    linespacing : float
        Line spacing
    backgroundcolor : str | tuple
        Background color
    margin : dict
        Margin around the text. The keys are one of
        ``['t', 'b', 'l', 'r']`` and ``units``. The units are
        one of ``['pt', 'lines', 'in']``. The *units* default
        to ``pt`` and the other keys to ``0``. Not all text
        themeables support margin parameters and other than the
        ``units``, only some of the other keys may apply.
    kwargs : dict
        Parameters recognised by :class:`matplotlib.text.Text`

    Notes
    -----
    :class:`element_text` will accept parameters that conform to the
    **ggplot2** *element_text* API, but it is preferable the
    **Matplotlib** based API described above.
    """

    def __init__(
        self,
        family: Optional[str | list[str]] = None,
        style: Optional[str] = None,
        weight: Optional[int | str] = None,
        color: Optional[str | TupleFloat3 | TupleFloat4] = None,
        size: Optional[float] = None,
        ha: Optional[Literal["center", "left", "right"]] = None,
        va: Optional[Literal["center", "top", "bottom", "baseline"]] = None,
        rotation: Optional[float] = None,
        linespacing: Optional[float] = None,
        backgroundcolor: Optional[str | TupleFloat3 | TupleFloat4] = None,
        margin: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ):
        # ggplot2 translation
        with suppress(KeyError):
            linespacing = kwargs.pop("lineheight")
        with suppress(KeyError):
            color = color or kwargs.pop("colour")
        with suppress(KeyError):
            _face = kwargs.pop("face")
            if _face == "plain":
                style = "normal"
            elif _face == "italic":
                style = "italic"
            elif _face == "bold":
                weight = "bold"
            elif _face == "bold.italic":
                style = "italic"
                weight = "bold"
        with suppress(KeyError):
            ha = self._translate_hjust(kwargs.pop("hjust"))
        with suppress(KeyError):
            va = self._translate_vjust(kwargs.pop("vjust"))
        with suppress(KeyError):
            rotation = kwargs.pop("angle")

        super().__init__()
        self.properties.update(**kwargs)

        if margin is not None:
            margin = Margin(self, **margin)  # type: ignore

        # Use the parameters that have been set
        names = (
            "backgroundcolor",
            "color",
            "family",
            "ha",
            "linespacing",
            "rotation",
            "size",
            "style",
            "va",
            "weight",
            "margin",
        )
        variables = locals()
        for name in names:
            if variables[name] is not None:
                self.properties[name] = variables[name]

    def setup(self, theme: Theme):
        """
        Setup the theme_element before drawing
        """
        if "margin" in self.properties:
            self.properties["margin"].theme = theme

    def _translate_hjust(
        self, just: float
    ) -> Literal["left", "right", "center"]:
        """
        Translate ggplot2 justification from [0, 1] to left, right, center.
        """
        if just == 0:
            return "left"
        elif just == 1:
            return "right"
        else:
            return "center"

    def _translate_vjust(
        self, just: float
    ) -> Literal["top", "bottom", "center"]:
        """
        Translate ggplot2 justification from [0, 1] to top, bottom, center.
        """
        if just == 0:
            return "bottom"
        elif just == 1:
            return "top"
        else:
            return "center"


class element_blank(element_base):
    """
    Theme element: Blank
    """

    def __init__(self):
        self.properties = {"visible": False}


@dataclass
class Margin:
    element: element_base
    theme: Optional[Theme] = None
    t: float = 0
    b: float = 0
    l: float = 0
    r: float = 0
    units: Literal["pt", "in", "lines", "fig"] = "pt"

    def __post_init__(self):
        if self.units in ("pts", "points", "px", "pixels"):
            self.units = "pt"
        elif self.units in ("in", "inch", "inches"):
            self.units = "in"
        elif self.units in ("line", "lines"):
            self.units = "lines"

    def __eq__(self, other: Any) -> bool:
        core = ("t", "b", "l", "r", "units")
        if self is other:
            return True

        if type(self) is not type(other):
            return False

        for attr in core:
            if getattr(self, attr) != getattr(other, attr):
                return False

        s_size = self.element.properties.get("size")
        o_size = other.element.properties.get("size")
        return s_size == o_size

    def get_as(
        self,
        loc: Literal["t", "b", "l", "r"],
        units: Literal["pt", "in", "lines", "fig"] = "pt",
    ) -> float:
        """
        Return key in given units
        """
        assert self.theme is not None
        dpi = 72
        # TODO: Get the inherited size. We need to consider the
        # themeables mro
        size: float = self.element.properties.get("size", 11)
        from_units = self.units
        to_units = units
        W: float
        H: float
        W, H = self.theme.themeables.property("figure_size")  # inches
        L = (W * dpi) if loc in "tb" else (H * dpi)  # pts

        functions: dict[str, Callable[[float], float]] = {
            "fig-in": lambda x: x * L / dpi,
            "fig-lines": lambda x: x * L / size,
            "fig-pt": lambda x: x * L,
            "in-fig": lambda x: x * dpi / L,
            "in-lines": lambda x: x * dpi / size,
            "in-pt": lambda x: x * dpi,
            "lines-fig": lambda x: x * size / L,
            "lines-in": lambda x: x * size / dpi,
            "lines-pt": lambda x: x * size,
            "pt-fig": lambda x: x / L,
            "pt-in": lambda x: x / dpi,
            "pt-lines": lambda x: x / size,
        }

        value: float = getattr(self, loc)
        if from_units != to_units:
            conversion = f"{self.units}-{units}"
            try:
                value = functions[conversion](value)
            except ZeroDivisionError:
                value = 0
        return value



from .elements import element_line, element_rect, element_text
from .theme import theme
from .theme_gray import theme_gray


class theme_bw(theme_gray):
    """
    White background with black gridlines

    Parameters
    ----------
    base_size : int, optional
        Base font size. All text sizes are a scaled versions of
        the base font size. Default is 11.
    base_family : str, optional
        Base font family.
    """

    def __init__(self, base_size=11, base_family=None):
        super().__init__(base_size, base_family)
        self += theme(
            axis_text=element_text(size=0.8 * base_size),
            legend_key=element_rect(color="#CCCCCC"),
            panel_background=element_rect(fill="white"),
            panel_border=element_rect(fill="None", color="#7f7f7f"),
            panel_grid_major=element_line(color="#E5E5E5", size=0.8),
            panel_grid_minor=element_line(color="#FAFAFA", size=1),
            strip_background=element_rect(
                fill="#CCCCCC", color="#7F7F7F", size=1
            ),
        )



import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError
from ..mapping.evaluation import after_stat
from .stat import stat


# Note: distribution should be a name from scipy.stat.distribution
@document
class stat_qq(stat):
    """
    Calculation for quantile-quantile plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    distribution : str (default: norm)
        Distribution or distribution function name. The default is
        *norm* for a normal probability plot. Objects that look enough
        like a stats.distributions instance (i.e. they have a ppf
        method) are also accepted. See :mod:`scipy stats <scipy.stats>`
        for available distributions.
    dparams : dict
        Distribution-specific shape parameters (shape parameters plus
        location and scale).
    quantiles : array_like, optional
        Probability points at which to calculate the theoretical
        quantile values. If provided, must be the same number as
        as the sample data points. The default is to use calculated
        theoretical points, use to ``alpha_beta`` control how
        these points are generated.
    alpha_beta : tuple
        Parameter values to use when calculating the quantiles.
        Default is :py:`(3/8, 3/8)`.

    See Also
    --------
    scipy.stats.mstats.plotting_positions : Uses ``alpha_beta``
        to calculate the quantiles.
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

         'theoretical'  # theoretical quantiles
         'sample'       # sample quantiles

    """
    REQUIRED_AES = {"sample"}
    DEFAULT_AES = {"x": after_stat("theoretical"), "y": after_stat("sample")}
    DEFAULT_PARAMS = {
        "geom": "qq",
        "position": "identity",
        "na_rm": False,
        "distribution": "norm",
        "dparams": {},
        "quantiles": None,
        "alpha_beta": (3 / 8, 3 / 8),
    }

    @classmethod
    def compute_group(cls, data, scales, **params):
        from scipy.stats.mstats import plotting_positions

        from .distributions import get_continuous_distribution

        sample = data["sample"].sort_values().values
        alpha, beta = params["alpha_beta"]
        quantiles = params["quantiles"]

        if quantiles is None:
            quantiles = plotting_positions(sample, alpha, beta)
        elif len(quantiles) != len(sample):
            raise PlotnineError(
                "The number of quantile values is not the same as "
                "the number of sample values."
            )

        quantiles = np.asarray(quantiles)
        cdist = get_continuous_distribution(params["distribution"])
        theoretical = cdist.ppf(quantiles, **params["dparams"])
        return pd.DataFrame({"sample": sample, "theoretical": theoretical})



print("import: 'plotnine'")
import plotnine

print("import: 'plotnine.coords'")
import plotnine.coords

print("import: 'plotnine.data'")
import plotnine.data

print("import: 'plotnine.facets'")
import plotnine.facets

print("import: 'plotnine.geoms'")
import plotnine.geoms

print("import: 'plotnine.guides'")
import plotnine.guides

print("import: 'plotnine.mapping'")
import plotnine.mapping

print("import: 'plotnine.positions'")
import plotnine.positions

print("import: 'plotnine.scales'")
import plotnine.scales

print("import: 'plotnine.stats'")
import plotnine.stats

print("import: 'plotnine.themes'")
import plotnine.themes




"""
Facets
"""
from .facet_grid import facet_grid
from .facet_null import facet_null
from .facet_wrap import facet_wrap
from .labelling import (
    as_labeller,
    label_both,
    label_context,
    label_value,
    labeller,
)

__all__ = [
    "facet_grid",
    "facet_null",
    "facet_wrap",
    "label_value",
    "label_both",
    "label_context",
    "labeller",
    "as_labeller",
]



from __future__ import annotations

import itertools
import types
import typing
from copy import copy, deepcopy

import numpy as np
import pandas as pd
import pandas.api.types as pdtypes

from ..exceptions import PlotnineError
from ..scales.scales import Scales
from ..utils import cross_join, match
from .strips import Strips

if typing.TYPE_CHECKING:
    from typing import Any, Literal, Optional

    import numpy.typing as npt
    from matplotlib.gridspec import GridSpec

    from plotnine.iapi import layout_details, panel_view
    from plotnine.typing import (
        Axes,
        Coord,
        EvalEnvironment,
        Figure,
        Ggplot,
        Layers,
        Layout,
        Scale,
        Theme,
    )


class facet:
    """
    Base class for all facets

    Parameters
    ----------
    scales : str in ``['fixed', 'free', 'free_x', 'free_y']``
        Whether ``x`` or ``y`` scales should be allowed (free)
        to vary according to the data along the rows or the
        columns. Default is ``'fixed'``.
    shrink : bool
        Whether to shrink the scales to the output of the
        statistics instead of the raw data. Default is ``True``.
    labeller : str | function
        How to label the facets. If it is a ``str``, it should
        be one of ``'label_value'`` ``'label_both'`` or
        ``'label_context'``. Default is ``'label_value'``
    as_table : bool
        If ``True``, the facets are laid out like a table with
        the highest values at the bottom-right. If ``False``
        the facets are laid out like a plot with the highest
        value a the top-right. Default it ``True``.
    drop : bool
        If ``True``, all factor levels not used in the data
        will automatically be dropped. If ``False``, all
        factor levels will be shown, regardless of whether
        or not they appear in the data. Default is ``True``.
    dir : str in ``['h', 'v']``
        Direction in which to layout the panels. ``h`` for
        horizontal and ``v`` for vertical.
    """

    #: number of columns
    ncol: int
    #: number of rows
    nrow: int
    as_table = True
    drop = True
    shrink = True
    #: Which axis scales are free
    free: dict[Literal["x", "y"], bool]
    #: A dict of parameters created depending on the data
    #: (Intended for extensions)
    params: dict[str, Any]
    # Theme object, automatically updated before drawing the plot
    theme: Theme
    # Figure object on which the facet panels are created
    figure: Figure
    # coord object, automatically updated before drawing the plot
    coordinates: Coord
    # layout object, automatically updated before drawing the plot
    layout: Layout
    # Axes
    axs: list[Axes]
    # The first and last axes according to how MPL creates them.
    # Used for labelling the x and y axes,
    first_ax: Axes
    last_ax: Axes
    # Number of facet variables along the horizontal axis
    num_vars_x = 0
    # Number of facet variables along the vertical axis
    num_vars_y = 0
    # ggplot object that the facet belongs to
    plot: Ggplot
    # Facet strips
    strips: Strips
    # Control the relative size of multiple facets
    # Use a subclass to change the default.
    # See: facet_grid for an example
    space: (
        Literal["fixed", "free", "free_x", "free_y"]
        | dict[Literal["x", "y"], list[int]]
    ) = "fixed"

    grid_spec: GridSpec

    def __init__(
        self,
        scales: Literal["fixed", "free", "free_x", "free_y"] = "fixed",
        shrink: bool = True,
        labeller: Literal[
            "label_value", "label_both", "label_context"
        ] = "label_value",
        as_table: bool = True,
        drop: bool = True,
        dir: Literal["h", "v"] = "h",
    ):
        from .labelling import as_labeller

        self.shrink = shrink
        self.labeller = as_labeller(labeller)
        self.as_table = as_table
        self.drop = drop
        self.dir = dir
        self.free = {
            "x": scales in ("free_x", "free"),
            "y": scales in ("free_y", "free"),
        }

    def __radd__(self, gg: Ggplot) -> Ggplot:
        """
        Add facet to ggplot object
        """
        gg.facet = copy(self)
        gg.facet.plot = gg
        return gg

    def set_properties(self, gg: Ggplot):
        """
        Copy required properties from ggplot object
        """
        self.axs = gg.axs
        self.coordinates = gg.coordinates
        self.figure = gg.figure
        self.layout = gg.layout
        self.theme = gg.theme
        self.strips = Strips.from_facet(self)

    def setup_data(self, data: list[pd.DataFrame]) -> list[pd.DataFrame]:
        """
        Allow the facet to manipulate the data

        Parameters
        ----------
        data : list of dataframes
            Data for each of the layers

        Returns
        -------
        data : list of dataframes
            Data for each of the layers

        Notes
        -----
        This method will be called after :meth:`setup_params`,
        therefore the `params` property will be set.
        """
        return data

    def setup_params(self, data: list[pd.DataFrame]):
        """
        Create facet parameters

        Parameters
        ----------
        data : list of dataframes
            Plot data and data for the layers
        """
        self.params = {}

    def init_scales(
        self,
        layout: pd.DataFrame,
        x_scale: Optional[Scale] = None,
        y_scale: Optional[Scale] = None,
    ) -> types.SimpleNamespace:
        scales = types.SimpleNamespace()

        if x_scale is not None:
            n = layout["SCALE_X"].max()
            scales.x = Scales([x_scale.clone() for i in range(n)])

        if y_scale is not None:
            n = layout["SCALE_Y"].max()
            scales.y = Scales([y_scale.clone() for i in range(n)])

        return scales

    def map(self, data: pd.DataFrame, layout: pd.DataFrame) -> pd.DataFrame:
        """
        Assign a data points to panels

        Parameters
        ----------
        data : DataFrame
            Data for a layer
        layout : DataFrame
            As returned by self.compute_layout

        Returns
        -------
        data : DataFrame
            Data with all points mapped to the panels
            on which they will be plotted.
        """
        msg = "{} should implement this method."
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def compute_layout(
        self,
        data: list[pd.DataFrame],
    ) -> pd.DataFrame:
        """
        Compute layout

        Parameters
        ----------
        data : Dataframes
            Dataframe for a each layer
        """
        msg = "{} should implement this method."
        raise NotImplementedError(msg.format(self.__class__.__name__))

    def finish_data(self, data: pd.DataFrame, layout: Layout) -> pd.DataFrame:
        """
        Modify data before it is drawn out by the geom

        The default is to return the data without modification.
        Subclasses should override this method as the require.

        Parameters
        ----------
        data : DataFrame
            A single layer's data.
        layout : Layout
            Layout

        Returns
        -------
        data : DataFrame
            Modified layer data
        """
        return data

    def train_position_scales(self, layout: Layout, layers: Layers) -> facet:
        """
        Compute ranges for the x and y scales
        """
        _layout = layout.layout
        panel_scales_x = layout.panel_scales_x
        panel_scales_y = layout.panel_scales_y

        # loop over each layer, training x and y scales in turn
        for layer in layers:
            data = layer.data
            match_id = match(data["PANEL"], _layout["PANEL"])
            if panel_scales_x:
                x_vars = list(
                    set(panel_scales_x[0].aesthetics) & set(data.columns)
                )
                # the scale index for each data point
                SCALE_X = _layout["SCALE_X"].iloc[match_id].tolist()
                panel_scales_x.train(data, x_vars, SCALE_X)

            if panel_scales_y:
                y_vars = list(
                    set(panel_scales_y[0].aesthetics) & set(data.columns)
                )
                # the scale index for each data point
                SCALE_Y = _layout["SCALE_Y"].iloc[match_id].tolist()
                panel_scales_y.train(data, y_vars, SCALE_Y)

        return self

    def make_ax_strips(self, layout_info: layout_details, ax: Axes) -> Strips:
        """
        Create strips for the facet

        Parameters
        ----------
        layout_info : dict-like
            Layout information. Row from the layout table

        ax : axes
            Axes to label
        """
        return Strips()

    def set_limits_breaks_and_labels(self, panel_params: panel_view, ax: Axes):
        """
        Add limits, breaks and labels to the axes

        Parameters
        ----------
        ranges : dict-like
            range information for the axes
        ax : Axes
            Axes
        """
        from .._mpl.ticker import MyFixedFormatter

        def _inf_to_none(
            t: tuple[float, float]
        ) -> tuple[float | None, float | None]:
            """
            Replace infinities with None
            """
            a = t[0] if np.isfinite(t[0]) else None
            b = t[1] if np.isfinite(t[1]) else None
            return (a, b)

        # limits
        ax.set_xlim(_inf_to_none(panel_params.x.range))
        ax.set_ylim(_inf_to_none(panel_params.y.range))

        if typing.TYPE_CHECKING:
            assert callable(ax.set_xticks)
            assert callable(ax.set_yticks)

        # breaks, labels
        ax.set_xticks(panel_params.x.breaks, panel_params.x.labels)
        ax.set_yticks(panel_params.y.breaks, panel_params.y.labels)

        # minor breaks
        ax.set_xticks(panel_params.x.minor_breaks, minor=True)
        ax.set_yticks(panel_params.y.minor_breaks, minor=True)

        # When you manually set the tick labels MPL changes the locator
        # so that it no longer reports the x & y positions
        # Fixes https://github.com/has2k1/plotnine/issues/187
        ax.xaxis.set_major_formatter(MyFixedFormatter(panel_params.x.labels))
        ax.yaxis.set_major_formatter(MyFixedFormatter(panel_params.y.labels))

        _property = self.theme.themeables.property
        margin = _property("axis_text_x", "margin")
        pad_x = margin.get_as("t", "pt")

        margin = _property("axis_text_y", "margin")
        pad_y = margin.get_as("r", "pt")

        ax.tick_params(axis="x", which="major", pad=pad_x)
        ax.tick_params(axis="y", which="major", pad=pad_y)

    def __deepcopy__(self, memo: dict[Any, Any]) -> facet:
        """
        Deep copy without copying the dataframe and environment
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        # don't make a deepcopy of the figure & the axes
        shallow = {"figure", "axs", "first_ax", "last_ax"}
        for key, item in old.items():
            if key in shallow:
                new[key] = old[key]
                memo[id(new[key])] = new[key]
            else:
                new[key] = deepcopy(old[key], memo)

        return result

    def _create_subplots(
        self, fig: Figure, layout: pd.DataFrame
    ) -> list[Axes]:
        """
        Create suplots and return axs
        """
        from matplotlib.gridspec import GridSpec

        num_panels = len(layout)
        axsarr = np.empty((self.nrow, self.ncol), dtype=object)
        space = self.space
        default_space: dict[Literal["x", "y"], list[int]] = {
            "x": [1 for x in range(self.ncol)],
            "y": [1 for x in range(self.nrow)],
        }

        if isinstance(space, str):
            if space == "fixed":
                space = default_space
            # TODO: Implement 'free', 'free_x' & 'free_y'
            else:
                space = default_space
        elif isinstance(space, dict):
            if "x" not in space:
                space["x"] = default_space["x"]
            if "y" not in space:
                space["y"] = default_space["y"]

        if len(space["x"]) != self.ncol:
            raise ValueError(
                "The number of x-ratios for the facet space sizes "
                "should match the number of columns."
            )

        if len(space["y"]) != self.nrow:
            raise ValueError(
                "The number of y-ratios for the facet space sizes "
                "should match the number of rows."
            )

        gs = GridSpec(
            self.nrow,
            self.ncol,
            height_ratios=space["y"],
            width_ratios=space["x"],
        )
        self.grid_spec = gs

        # Create axes
        i = 1
        for row in range(self.nrow):
            for col in range(self.ncol):
                axsarr[row, col] = fig.add_subplot(gs[i - 1])
                i += 1

        # Rearrange axes
        # They are ordered to match the positions in the layout table
        if self.dir == "h":
            order: Literal["C", "F"] = "C"
            if not self.as_table:
                axsarr = axsarr[::-1]
        elif self.dir == "v":
            order = "F"
            if not self.as_table:
                axsarr = np.array([row[::-1] for row in axsarr])
        else:
            raise ValueError(f"Bad value `dir='{self.dir}'` for direction")

        axs = axsarr.ravel(order)

        # Delete unused axes
        for ax in axs[num_panels:]:
            fig.delaxes(ax)
        axs = axs[:num_panels]
        return list(axs)

    def make_axes(
        self, figure: Figure, layout: pd.DataFrame, coordinates: Coord
    ) -> list[Axes]:
        """
        Create and return Matplotlib axes
        """
        axs = self._create_subplots(figure, layout)

        # Used for labelling the x and y axes, the first and
        # last axes according to how MPL creates them.
        self.first_ax = figure.axes[0]
        self.last_ax = figure.axes[-1]
        self.figure = figure
        self.axs = axs
        return axs

    def _aspect_ratio(self) -> Optional[float]:
        """
        Return the aspect_ratio
        """
        aspect_ratio = self.theme.themeables.property("aspect_ratio")
        if aspect_ratio == "auto":
            # If the panels have different limits the coordinates
            # cannot compute a common aspect ratio
            if not self.free["x"] and not self.free["y"]:
                aspect_ratio = self.coordinates.aspect(
                    self.layout.panel_params[0]
                )
            else:
                aspect_ratio = None

        return aspect_ratio


def combine_vars(
    data: list[pd.DataFrame],
    environment: EvalEnvironment,
    vars: list[str],
    drop: bool = True,
) -> pd.DataFrame:
    """
    Generate all combinations of data needed for facetting

    The first data frame in the list should be the default data
    for the plot. Other data frames in the list are ones that are
    added to the layers.
    """
    if len(vars) == 0:
        return pd.DataFrame()

    # For each layer, compute the facet values
    values = [
        eval_facet_vars(df, vars, environment) for df in data if df is not None
    ]

    # Form the base data frame which contains all combinations
    # of facetting variables that appear in the data
    has_all = [x.shape[1] == len(vars) for x in values]
    if not any(has_all):
        raise PlotnineError(
            "At least one layer must contain all variables "
            "used for facetting"
        )
    base = pd.concat([x for i, x in enumerate(values) if has_all[i]], axis=0)
    base = base.drop_duplicates()

    if not drop:
        base = unique_combs(base)

    # sorts according to order of factor levels
    base = base.sort_values(base.columns.tolist())

    # Systematically add on missing combinations
    for i, value in enumerate(values):
        if has_all[i] or len(value.columns) == 0:
            continue
        old = base.loc[:, list(base.columns.difference(value.columns))]
        new = value.loc[
            :, list(base.columns.intersection(value.columns))
        ].drop_duplicates()

        if not drop:
            new = unique_combs(new)

        base = pd.concat([base, cross_join(old, new)], ignore_index=True)

    if len(base) == 0:
        raise PlotnineError("Faceting variables must have at least one value")

    base = base.reset_index(drop=True)
    return base


def unique_combs(df: pd.DataFrame) -> pd.DataFrame:
    """
    Generate all possible combinations of the values in the columns
    """

    def _unique(s: pd.Series[Any]) -> npt.NDArray[Any] | pd.Index:
        if isinstance(s.dtype, pdtypes.CategoricalDtype):
            return s.cat.categories
        return s.unique()

    # List of unique values from every column
    lst = (_unique(x) for _, x in df.items())
    rows = list(itertools.product(*lst))
    _df = pd.DataFrame(rows, columns=df.columns)

    # preserve the column dtypes
    for col in df:
        t = df[col].dtype
        _df[col] = _df[col].astype(t, copy=False)  # pyright: ignore
    return _df


def layout_null() -> pd.DataFrame:
    """
    Layout Null
    """
    layout = pd.DataFrame(
        {
            "PANEL": [1],
            "ROW": 1,
            "COL": 1,
            "SCALE_X": 1,
            "SCALE_Y": 1,
            "AXIS_X": True,
            "AXIS_Y": True,
        }
    )
    return layout


def add_missing_facets(
    data: pd.DataFrame,
    layout: pd.DataFrame,
    vars: list[str],
    facet_vals: pd.DataFrame,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Add missing facets
    """
    # When in a dataframe some layer does not have all
    # the facet variables, add the missing facet variables
    # and create new data where the points(duplicates) are
    # present in all the facets
    missing_facets = list(set(vars) - set(facet_vals.columns.tolist()))
    if missing_facets:
        to_add = layout.loc[:, missing_facets].drop_duplicates()
        to_add.reset_index(drop=True, inplace=True)

        # a point for each facet, [0, 1, ..., n-1, 0, 1, ..., n-1, ...]
        data_rep = np.tile(np.arange(len(data)), len(to_add))
        # a facet for each point, [0, 0, 0, 1, 1, 1, ... n-1, n-1, n-1]
        facet_rep = np.repeat(np.arange(len(to_add)), len(data))

        data = data.iloc[data_rep, :].reset_index(drop=True)
        facet_vals = facet_vals.iloc[data_rep, :].reset_index(drop=True)
        to_add = to_add.iloc[facet_rep, :].reset_index(drop=True)
        facet_vals = pd.concat(
            [facet_vals, to_add], axis=1, ignore_index=False
        )

    return data, facet_vals


def eval_facet_vars(
    data: pd.DataFrame, vars: list[str], env: EvalEnvironment
) -> pd.DataFrame:
    """
    Evaluate facet variables

    Parameters
    ----------
    data : DataFrame
        Factet dataframe
    vars : list
        Facet variables
    env : environment
        Plot environment

    Returns
    -------
    facet_vals : DataFrame
        Facet values that correspond to the specified
        variables.
    """

    # To allow expressions in facet formula
    def I(value: Any) -> Any:
        return value

    env = env.with_outer_namespace({"I": I})
    facet_vals = pd.DataFrame(index=data.index)

    for name in vars:
        if name in data:
            # This is a limited solution. If a keyword is
            # part of an expression it will fail in the
            # else statement below
            res = data[name]
        elif str.isidentifier(name):
            # All other non-statements
            continue
        else:
            # Statements
            try:
                res = env.eval(name, inner_namespace=data)
            except NameError:
                continue
        facet_vals[name] = res

    return facet_vals



from ..doctools import document
from .geom_area import geom_area


@document
class geom_density(geom_area):
    """
    Smooth density estimate

    {usage}

    Parameters
    ----------
    {common_parameters}

    See Also
    --------
    plotnine.geoms.geom_ribbon
    """

    DEFAULT_AES = dict(
        geom_area.DEFAULT_AES, **{"color": "black", "fill": None, "weight": 1}
    )

    DEFAULT_PARAMS = dict(
        geom_area.DEFAULT_PARAMS, **{"stat": "density", "position": "identity"}
    )



import numpy as np
import pandas as pd
import pytest
from mizani.transforms import trans_new

from plotnine import (
    aes,
    coord_fixed,
    coord_flip,
    coord_trans,
    geom_bar,
    geom_line,
    ggplot,
    xlim,
)

n = 10  # Some even number greater than 2

# ladder: 0 1 times, 1 2 times, 2 3 times, ...
df = pd.DataFrame(
    {
        "x": np.repeat(range(n + 1), range(n + 1)),
        "z": np.repeat(range(n // 2), range(3, n * 2, 4)),
    }
)

p = ggplot(df, aes("x")) + geom_bar(aes(fill="factor(z)"), show_legend=False)


def test_coord_flip():
    assert p + coord_flip() == "coord_flip"


def test_coord_fixed():
    assert p + coord_fixed(0.5) == "coord_fixed"


def test_coord_trans():
    double_trans = trans_new("double", np.square, np.sqrt)
    # Warns probably because of a bad value around the left
    # edge of the domain.
    with pytest.warns(RuntimeWarning):
        assert p + coord_trans(y=double_trans) == "coord_trans"


def test_coord_trans_reverse():
    # coord trans can reverse continous and discrete data
    p = (
        ggplot(df, aes("factor(x)"))
        + geom_bar(aes(fill="factor(z)"), show_legend=False)
        + coord_trans(x="reverse", y="reverse")
    )
    assert p == "coord_trans_reverse"


def test_coord_trans_backtransforms():
    df = pd.DataFrame({"x": [-np.inf, np.inf], "y": [1, 2]})
    p = (
        ggplot(df, aes("x", "y"))
        + geom_line(size=2)
        + xlim(1, 2)
        + coord_trans(x="log10")
    )
    assert p == "coord_trans_backtransform"



import pandas as pd

from plotnine import aes, geom_crossbar, ggplot

n = 4
df = pd.DataFrame(
    {
        "x": [1] * n,
        "ymin": range(1, 2 * n + 1, 2),
        "y": [i + 0.1 + i / 10 for i in range(1, 2 * n + 1, 2)],
        "ymax": range(2, 2 * n + 2, 2),
        "z": range(n),
    }
)


def test_aesthetics():
    p = (
        ggplot(df, aes(y="y", ymin="ymin", ymax="ymax"))
        + geom_crossbar(aes("x"), size=2)
        + geom_crossbar(aes("x+1", alpha="z"), fill="green", width=0.2, size=2)
        + geom_crossbar(aes("x+2", linetype="factor(z)"), size=2)
        + geom_crossbar(aes("x+3", color="factor(z)"), size=2)
        + geom_crossbar(aes("x+4", size="z"))
    )

    assert p == "aesthetics"



import numpy as np
import pandas as pd
import pytest

from plotnine import aes, coord_flip, geom_sina, geom_violin, ggplot

n = 50
random_state = np.random.RandomState(123)
uni = random_state.normal(5, 0.25, n)
bi = np.hstack(
    [random_state.normal(4, 0.25, n), random_state.normal(6, 0.25, n)]
)
tri = np.hstack(
    [
        random_state.normal(4, 0.125, n),
        random_state.normal(5, 0.125, n),
        random_state.normal(6, 0.125, n),
    ]
)

cats = ["uni", "bi", "tri"]

df = pd.DataFrame(
    {
        "dist": pd.Categorical(
            np.repeat(cats, [len(uni), len(bi), len(tri)]), categories=cats
        ),
        "value": np.hstack([uni, bi, tri]),
    }
)


@pytest.mark.skip("Fix the limits")
def test_scale_area():
    p = (
        ggplot(df, aes("dist", "value"))
        + geom_violin(scale="area")
        + geom_sina(scale="area", random_state=123)
    )

    assert p == "scale_area"


@pytest.mark.skip("Fix the limits")
def test_scale_count():
    p = (
        ggplot(df, aes("dist", "value"))
        + geom_violin(scale="count")
        + geom_sina(scale="count", random_state=123)
    )

    assert p == "scale_count"


@pytest.mark.skip("Fix the limits")
def test_scale_area_coordflip():
    p = (
        ggplot(df, aes("dist", "value"))
        + geom_violin(scale="area")
        + geom_sina(scale="area", random_state=123)
        + coord_flip()
    )

    assert p == "scale_area+coord_flip"


@pytest.mark.skip("Fix the limits")
def test_method_counts():
    p = (
        ggplot(df, aes("dist", "value"))
        + geom_violin()
        + geom_sina(method="counts", random_state=123)
    )

    assert p == "method_counts"



from __future__ import annotations

import warnings
from textwrap import dedent
from typing import Optional, Type, Union


# Show the warnings on one line, leaving out any code makes the
# message clear
def warning_format(
    message: Union[Warning, str],
    category: Type[Warning],
    filename: str,
    lineno: int,
    line: Optional[str] = None,
) -> str:
    """
    Format for plotnine warnings
    """
    fmt = "{}:{}: {}: {}\n".format
    return fmt(filename, lineno, category.__name__, message)


warnings.formatwarning = warning_format


class PlotnineError(Exception):
    """
    Exception for ggplot errors
    """

    def __init__(self, *args: str):
        args = tuple(dedent(arg) for arg in args)
        self.message = " ".join(args)

    def __str__(self) -> str:
        """
        Error Message
        """
        return repr(self.message)


class PlotnineWarning(UserWarning):
    """
    Warnings for ggplot inconsistencies
    """

    pass



from __future__ import annotations

import typing

from matplotlib.text import Text

from .patches import SFancyBboxPatch
from .utils import bbox_in_axes_space

if typing.TYPE_CHECKING:
    from matplotlib.backend_bases import RendererBase

    from plotnine.iapi import strip_draw_info


class SText(Text):
    """
    Strip Text
    """

    draw_info: strip_draw_info
    spatch: SFancyBboxPatch

    def __init__(self, info: strip_draw_info):
        kwargs = {
            "ha": info.ha,
            "va": info.va,
            "rotation": info.rotation,
            "transform": info.ax.transAxes,
            "clip_on": False,
            "zorder": 3.3,
        }

        super().__init__(
            info.x, info.y, info.label, **kwargs  # pyright: ignore
        )
        self.draw_info = info
        self.spatch = SFancyBboxPatch(self)

    def draw(self, renderer: RendererBase):
        info = self.draw_info
        # "fill up" spatch to contain the text
        self.spatch.update_position_size(renderer)

        # Get bbox of spatch in transAxes space
        patch_bbox = bbox_in_axes_space(self.spatch, info.ax, renderer)

        # Align patch across the edge of the panel
        if info.position == "top":
            l, b, w, h = info.x, info.y, info.box_width, patch_bbox.height
            b = b + patch_bbox.height * info.strip_align
        else:  # "right"
            l, b, w, h = info.x, info.y, patch_bbox.width, info.box_height
            l = l + patch_bbox.width * info.strip_align

        self.spatch.set_bounds(l, b, w, h)
        self.spatch.set_transform(info.ax.transAxes)
        self.spatch.set_mutation_scale(0)

        # Put text in center of patch
        self._x = l + w / 2
        self._y = b + h / 2
        self.set_horizontalalignment("center")  # right-strip
        self.set_verticalalignment("center")  # top-strip

        # Draw spatch
        self.spatch.draw(renderer)
        return super().draw(renderer)



from __future__ import annotations

import typing
from collections.abc import Sequence
from copy import deepcopy
from itertools import chain
from pathlib import Path
from types import SimpleNamespace as NS
from typing import Any, Dict, Iterable, Optional, Union
from warnings import warn

import pandas as pd

from .coords import coord_cartesian
from .exceptions import PlotnineError, PlotnineWarning
from .facets import facet_null
from .facets.layout import Layout
from .geoms.geom_blank import geom_blank
from .guides.guides import guides
from .iapi import mpl_save_view
from .layer import Layers
from .mapping.aes import aes, make_labels
from .options import get_option
from .scales.scales import Scales
from .themes.theme import theme, theme_get
from .utils import (
    from_inches,
    get_ipython,
    is_data_like,
    order_as_data_mapping,
    to_inches,
    ungroup,
)

if typing.TYPE_CHECKING:
    from plotnine.typing import (
        Axes,
        Coord,
        DataLike,
        EvalEnvironment,
        Facet,
        Figure,
        Layer,
        PlotAddable,
        Theme,
        Watermark,
    )


class ggplot:
    """
    Create a new ggplot object

    Parameters
    ----------
    data : dataframe
        Default data for plot. Every layer that does not
        have data of its own will use this one.
    mapping : aes
        Default aesthetics mapping for the plot. These will be used
        by all layers unless specifically overridden.
    environment : ~patsy.Eval.EvalEnvironment
        If a variable defined in the aesthetic mapping is not
        found in the data, ggplot will look for it in this
        namespace. It defaults to using the environment/namespace.
        in which `ggplot()` is called.
    """

    figure: Figure
    axs: list[Axes]
    theme: Theme
    facet: Facet
    coordinates: Coord

    def __init__(
        self,
        data: DataLike | None = None,
        mapping: aes | None = None,
        environment: EvalEnvironment | None = None,
    ):
        from patsy.eval import EvalEnvironment

        # Allow some sloppiness
        data, mapping = order_as_data_mapping(data, mapping)
        self.data = data
        self.mapping = mapping if mapping is not None else aes()
        self.facet = facet_null()
        self.labels = make_labels(self.mapping)
        self.layers = Layers()
        self.guides = guides()
        self.scales = Scales()
        self.theme = theme_get()
        self.coordinates: Coord = coord_cartesian()
        self.environment = environment or EvalEnvironment.capture(1)
        self.layout = Layout()
        self.watermarks: list[Watermark] = []

        # build artefacts
        self._build_objs = NS()

    def __str__(self) -> str:
        """
        Print/show the plot
        """
        self.draw(show=True)

        # Return and empty string so that print(p) is "pretty"
        return ""

    def __repr__(self) -> str:
        """
        Print/show the plot
        """
        figure = self.draw(show=True)

        dpi = figure.get_dpi()
        W = int(figure.get_figwidth() * dpi)
        H = int(figure.get_figheight() * dpi)

        return f"<Figure Size: ({W} x {H})>"

    def __deepcopy__(self, memo: dict[Any, Any]) -> ggplot:
        """
        Deep copy without copying the dataframe and environment
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        # don't make a deepcopy of data, or environment
        shallow = {
            "data",
            "environment",
            "figure",
            "_build_objs",
        }
        for key, item in old.items():
            if key in shallow:
                new[key] = old[key]
                memo[id(new[key])] = new[key]
            else:
                new[key] = deepcopy(old[key], memo)

        return result

    def __iadd__(
        self, other: PlotAddable | list[PlotAddable] | None
    ) -> ggplot:
        """
        Add other to ggplot object

        Parameters
        ----------
        other : object or Sequence
            Either an object that knows how to "radd"
            itself to a ggplot, or a list of such objects.
        """
        if isinstance(other, Sequence):
            for item in other:
                item.__radd__(self)
            return self
        elif other is None:
            return self
        else:
            return other.__radd__(self)

    def __add__(self, other: PlotAddable | list[PlotAddable] | None) -> ggplot:
        """
        Add to ggplot from a list

        Parameters
        ----------
        other : object or Sequence
            Either an object that knows how to "radd"
            itself to a ggplot, or a list of such objects.
        """
        self = deepcopy(self)
        return self.__iadd__(other)

    def __rrshift__(self, other: DataLike) -> ggplot:
        """
        Overload the >> operator to receive a dataframe
        """
        other = ungroup(other)
        if is_data_like(other):
            if self.data is None:
                self.data = other
            else:
                raise PlotnineError("`>>` failed, ggplot object has data.")
        else:
            msg = "Unknown type of data -- {!r}"
            raise TypeError(msg.format(type(other)))
        return self

    def draw(self, show: bool = False) -> Figure:
        """
        Render the complete plot

        Parameters
        ----------
        show : bool (default: False)
            Whether to show the plot.

        Returns
        -------
        fig : ~matplotlib.figure.Figure
            Matplotlib figure
        """
        from ._mpl.layout_engine import PlotnineLayoutEngine

        # Do not draw if drawn already.
        # This prevents a needless error when reusing
        # figure & axes in the jupyter notebook.
        if hasattr(self, "figure"):
            return self.figure

        # Prevent against any modifications to the users
        # ggplot object. Do the copy here as we may/may not
        # assign a default theme
        self = deepcopy(self)
        with plot_context(self, show=show):
            self._build()

            # setup
            figure, axs = self._create_figure()
            self._setup_parameters()
            self.theme.setup()
            self.facet.strips.generate()

            # Drawing
            self._draw_layers()
            self._draw_breaks_and_labels()
            self._draw_legend()
            self._draw_figure_texts()
            self._draw_watermarks()

            # Artist object theming
            self.theme.apply()
            figure.set_layout_engine(PlotnineLayoutEngine(self))

        return self.figure

    def _draw_using_figure(self, figure: Figure, axs: list[Axes]) -> ggplot:
        """
        Draw onto already created figure and axes

        This is can be used to draw animation frames,
        or inset plots. It is intended to be used
        after the key plot has been drawn.

        Parameters
        ----------
        figure : ~matplotlib.figure.Figure
            Matplotlib figure
        axs : array_like
            Array of Axes onto which to draw the plots
        """
        from ._mpl.layout_engine import PlotnineLayoutEngine

        self = deepcopy(self)
        self.figure = figure
        self.axs = axs
        with plot_context(self):
            self._build()

            # setup
            self._setup_parameters()
            self.theme.setup()
            self.facet.strips.generate()

            # drawing
            self._draw_layers()
            self._draw_breaks_and_labels()
            self._draw_legend()

            # artist theming
            self.theme.apply()
            figure.set_layout_engine(PlotnineLayoutEngine(self))

        return self

    def _build(self):
        """
        Build ggplot for rendering.

        Notes
        -----
        This method modifies the ggplot object. The caller is
        responsible for making a copy and using that to make
        the method call.
        """
        if not self.layers:
            self += geom_blank()

        layers = self._build_objs.layers = self.layers
        scales = self._build_objs.scales = self.scales
        layout = self._build_objs.layout = self.layout

        # Update the label information for the plot
        layers.update_labels(self)

        # Give each layer a copy of the data, the mappings and
        # the execution environment
        layers.setup(self)

        # Initialise panels, add extra data for margins & missing
        # facetting variables, and add on a PANEL variable to data
        layout.setup(layers, self)

        # Compute aesthetics to produce data with generalised
        # variable names
        layers.compute_aesthetics(self)

        # Transform data using all scales
        layers.transform(scales)

        # Make sure missing (but required) aesthetics are added
        scales.add_missing(("x", "y"))

        # Map and train positions so that statistics have access
        # to ranges and all positions are numeric
        layout.train_position(layers, scales)
        layout.map_position(layers)

        # Apply and map statistics
        layers.compute_statistic(layout)
        layers.map_statistic(self)

        # Prepare data in geoms
        # e.g. from y and width to ymin and ymax
        layers.setup_data()

        # Apply position adjustments
        layers.compute_position(layout)

        # Reset position scales, then re-train and map.  This
        # ensures that facets have control over the range of
        # a plot.
        layout.reset_position_scales()
        layout.train_position(layers, scales)
        layout.map_position(layers)

        # Train and map non-position scales
        npscales = scales.non_position_scales()
        if len(npscales):
            layers.train(npscales)
            layers.map(npscales)

        # Train coordinate system
        layout.setup_panel_params(self.coordinates)

        # fill in the defaults
        layers.use_defaults()

        # Allow stats to modify the layer data
        layers.finish_statistics()

        # Allow layout to modify data before rendering
        layout.finish_data(layers)

    def _setup_parameters(self):
        """
        Set facet properties
        """
        # facet
        self.facet.set_properties(self)
        # layout
        self.layout.axs = self.axs
        # theme
        self.theme.figure = self.figure
        self.theme.axs = self.axs

    def _create_figure(self) -> tuple[Figure, list[Axes]]:
        """
        Create Matplotlib figure and axes
        """
        import matplotlib.pyplot as plt

        # Good for development
        if get_option("close_all_figures"):
            plt.close("all")

        figure: Figure = plt.figure()  # pyright: ignore
        axs = self.facet.make_axes(
            figure, self.layout.layout, self.coordinates
        )

        self.figure = figure
        self.axs = axs
        return figure, axs

    def _draw_layers(self):
        """
        Draw the main plot(s) onto the axes.
        """
        # Draw the geoms
        self.layers.draw(self.layout, self.coordinates)

    def _draw_breaks_and_labels(self):
        """
        Draw breaks and labels
        """
        # 1. Draw facet labels a.k.a strip text
        # 2. Decorate the axes
        #      - xaxis & yaxis breaks, labels, limits, ...
        #
        # pidx is the panel index (location left to right, top to bottom)
        self.facet.strips.draw()
        for layout_info in self.layout.get_details():
            pidx = layout_info.panel_index
            ax = self.axs[pidx]
            panel_params = self.layout.panel_params[pidx]
            self.facet.set_limits_breaks_and_labels(panel_params, ax)

            # Remove unnecessary ticks and labels
            if not layout_info.axis_x:
                ax.xaxis.set_tick_params(
                    which="both", bottom=False, labelbottom=False
                )
            if not layout_info.axis_y:
                ax.yaxis.set_tick_params(
                    which="both", left=False, labelleft=False
                )

            if layout_info.axis_x:
                ax.xaxis.set_tick_params(which="both", bottom=True)
            if layout_info.axis_y:
                ax.yaxis.set_tick_params(which="both", left=True)

    def _draw_legend(self):
        """
        Draw legend onto the figure
        """
        from matplotlib.offsetbox import AnchoredOffsetbox

        legend_box = self.guides.build(self)
        if not legend_box:
            return

        anchored_box = AnchoredOffsetbox(
            loc="center",
            child=legend_box,
            pad=0.0,
            frameon=False,
            prop={"size": 0, "stretch": 0},
            bbox_to_anchor=(0, 0),
            bbox_transform=self.figure.transFigure,
            borderpad=0.0,
        )

        anchored_box.set_zorder(90.1)
        anchored_box.set_in_layout(True)

        self.theme._targets["legend_background"] = anchored_box
        self.figure.add_artist(anchored_box)

    def _draw_figure_texts(self):
        """
        Draw title, x label, y label and caption onto the figure
        """
        figure = self.figure
        theme = self.theme

        title = self.labels.get("title", "")
        caption = self.labels.get("caption", "")
        subtitle = self.labels.get("subtitle", "")

        # Get the axis labels (default or specified by user)
        # and let the coordinate modify them e.g. flip
        labels = self.coordinates.labels(
            self.layout.set_xy_labels(self.labels)
        )

        # The locations are handled by the layout manager
        text_title = figure.text(0, 0, title)
        text_caption = figure.text(0, 0, caption)
        text_subtitle = figure.text(0, 0, subtitle)
        text_x = figure.text(0, 0, labels.x)
        text_y = figure.text(0, 0, labels.y)

        theme._targets["plot_title"] = text_title
        theme._targets["plot_caption"] = text_caption
        theme._targets["plot_subtitle"] = text_subtitle
        theme._targets["axis_title_x"] = text_x
        theme._targets["axis_title_y"] = text_y

    def _draw_watermarks(self):
        """
        Draw watermark onto figure
        """
        for wm in self.watermarks:
            wm.draw(self.figure)

    def _save_filename(self, ext: str) -> Path:
        """
        Make a filename for use by the save method

        Parameters
        ----------
        ext : str
            Extension e.g. png, pdf, ...
        """
        hash_token = abs(self.__hash__())
        return Path(f"plotnine-save-{hash_token}.{ext}")

    def _update_labels(self, layer: Layer):
        """
        Update label data for the ggplot

        Parameters
        ----------
        layer : layer
            New layer that has just been added to the ggplot
            object.
        """
        mapping = make_labels(layer.mapping)
        default = make_labels(layer.stat.DEFAULT_AES)
        mapping.add_defaults(default)
        self.labels.add_defaults(mapping)

    def save_helper(
        self: ggplot,
        filename: Optional[Union[str, Path]] = None,
        format: Optional[str] = None,
        path: Optional[str] = None,
        width: Optional[float] = None,
        height: Optional[float] = None,
        units: str = "in",
        dpi: Optional[float] = None,
        limitsize: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> mpl_save_view:
        """
        Create MPL figure that will be saved

        Notes
        -----
        This method has the same arguments as :meth:`ggplot.save`.
        Use it to get access to the figure that will be saved.
        """
        fig_kwargs: Dict[str, Any] = {"format": format, **kwargs}

        # filename, depends on the object
        if filename is None:
            ext = format if format else "pdf"
            filename = self._save_filename(ext)

        if path:
            filename = Path(path) / filename

        fig_kwargs["fname"] = filename

        # Preserve the users object
        self = deepcopy(self)

        # The figure size should be known by the theme
        if width is not None and height is not None:
            width = to_inches(width, units)
            height = to_inches(height, units)
            self += theme(figure_size=(width, height))
        elif (
            width is None
            and height is not None
            or width is not None
            and height is None
        ):
            raise PlotnineError("You must specify both width and height")

        width, height = self.theme.themeables.property("figure_size")
        assert width is not None
        assert height is not None

        if limitsize and (width > 25 or height > 25):
            raise PlotnineError(
                f"Dimensions ({width=}, {height=}) exceed 25 inches "
                "(height and width are specified in inches/cm/mm, "
                "not pixels). If you are sure you want these "
                "dimensions, use 'limitsize=False'."
            )

        if verbose:
            _w = from_inches(width, units)
            _h = from_inches(height, units)
            warn(f"Saving {_w} x {_h} {units} image.", PlotnineWarning)
            warn(f"Filename: {filename}", PlotnineWarning)

        if dpi is not None:
            self.theme = self.theme + theme(dpi=dpi)

        figure = self.draw(show=False)
        return mpl_save_view(figure, fig_kwargs)

    def save(
        self,
        filename: Union[str, Path] | None = None,
        format: str | None = None,
        path: str = "",
        width: float | None = None,
        height: float | None = None,
        units: str = "in",
        dpi: float | None = None,
        limitsize: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ):
        """
        Save a ggplot object as an image file

        Parameters
        ----------
        filename : str | pathlib.Path, optional
            File name to write the plot to. If not specified, a name
            like “plotnine-save-<hash>.<format>” is used.
        format : str
            Image format to use, automatically extract from
            file name extension.
        path : str
            Path to save plot to (if you just want to set path and
            not filename).
        width : number, optional
            Width (defaults to value set by the theme). If specified
            the `height` must also be given.
        height : number, optional
            Height (defaults to value set by the theme). If specified
            the `width` must also be given.
        units : str
            Units for width and height when either one is explicitly
            specified (in, cm, or mm).
        dpi : float
            DPI to use for raster graphics. If None, defaults to using
            the `dpi` of theme, if none is set then a `dpi` of 100.
        limitsize : bool
            If ``True`` (the default), ggsave will not save images
            larger than 50x50 inches, to prevent the common error
            of specifying dimensions in pixels.
        verbose : bool
            If ``True``, print the saving information.
        kwargs : dict
            Additional arguments to pass to matplotlib `savefig()`.
        """
        sv = self.save_helper(
            filename=filename,
            format=format,
            path=path,
            width=width,
            height=height,
            units=units,
            dpi=dpi,
            limitsize=limitsize,
            verbose=verbose,
            **kwargs,
        )
        sv.figure.savefig(**sv.kwargs)


ggsave = ggplot.save


def save_as_pdf_pages(
    plots: Iterable[ggplot],
    filename: Optional[str | Path] = None,
    path: str | None = None,
    verbose: bool = True,
    **kwargs: Any,
):
    """
    Save multiple :class:`ggplot` objects to a PDF file, one per page.

    Parameters
    ----------
    plots : collection or generator of :class:`ggplot`
        Plot objects to write to file. `plots` may be either a
        collection such as a :py:class:`list` or :py:class:`set`:

        >>> base_plot = ggplot(…)
        >>> plots = [base_plot + ggtitle('%d of 3' % i) for i in range(1, 3)]
        >>> save_as_pdf_pages(plots)

        or, a generator that yields :class:`ggplot` objects:

        >>> def myplots():
        >>>     for i in range(1, 3):
        >>>         yield ggplot(…) + ggtitle('%d of 3' % i)
        >>> save_as_pdf_pages(myplots())

    filename : :py:class:`str`, optional
        File name to write the plot to. If not specified, a name
        like “plotnine-save-<hash>.pdf” is used.
    path : :py:class:`str`, optional
        Path to save plot to (if you just want to set path and
        not filename).
    verbose : :py:class:`bool`
        If ``True``, print the saving information.
    kwargs : :py:class:`dict`
        Additional arguments to pass to
        :py:meth:`matplotlib.figure.Figure.savefig`.

    Notes
    -----
    Using pandas' :meth:`~pandas.DataFrame.groupby` methods, tidy data
    can be “faceted” across pages:

    >>> from plotnine.data import mtcars
    >>> def facet_pages(column)
    >>>     base_plot = [
    >>>         aes(x='wt', y='mpg', label='name'),
    >>>         geom_text(),
    >>>         ]
    >>>     for label, group_data in mtcars.groupby(column):
    >>>         yield ggplot(group_data) + base_plot + ggtitle(label)
    >>> save_as_pdf_pages(facet_pages('cyl'))

    Unlike :meth:`ggplot.save`, :meth:`save_as_pdf_pages` does not
    process arguments for `height` or `width`. To set the figure size,
    add :class:`~plotnine.themes.themeable.figure_size` to the theme
    for some or all of the objects in `plots`:

    >>> plot = ggplot(…)
    >>> # The following are equivalent
    >>> plot.save('filename.pdf', height=6, width=8)
    >>> save_as_pdf_pages([plot + theme(figure_size=(8, 6))])
    """
    from matplotlib.backends.backend_pdf import PdfPages

    # as in ggplot.save()
    fig_kwargs = {"bbox_inches": "tight"}
    fig_kwargs.update(kwargs)

    # If plots is already an iterator, this is a no-op; otherwise
    # convert a list, etc. to an iterator
    plots = iter(plots)

    # filename, depends on the object
    if filename is None:
        # Take the first element from the iterator, store it, and
        # use it to generate a file name
        peek = [next(plots)]
        plots = chain(peek, plots)
        filename = peek[0]._save_filename("pdf")

    if path:
        filename = Path(path) / filename

    if verbose:
        warn(f"Filename: {filename}", PlotnineWarning)

    with PdfPages(filename) as pdf:
        # Re-add the first element to the iterator, if it was removed
        for plot in plots:
            fig = plot.draw()
            # Save as a page in the PDF file
            pdf.savefig(fig, **fig_kwargs)


class plot_context:
    """
    Context to setup the environment within with the plot is built

    Parameters
    ----------
    plot : ggplot
        ggplot object to be built within the context.
    show : bool (default: False)
        Whether to show (``plt.show()``) the plot before the context
        exits.
    """

    # Default to retina unless user chooses otherwise
    _IPYTHON_CONFIG: dict[str, dict[str, Any]] = {
        "InlineBackend": {
            "figure_format": "retina",
            "close_figures": True,
            "print_figure_kwargs": {"bbox_inches": None},
        }
    }
    _ip_config_inlinebackend: dict[str, Any] = {}

    def __init__(self, plot: ggplot, show: bool = False):
        self.plot = plot
        self.show = show

    def __enter__(self) -> plot_context:
        """
        Enclose in matplolib & pandas environments
        """
        import matplotlib as mpl

        self.plot.theme._targets = {}
        self.rc_context = mpl.rc_context(self.plot.theme.rcParams)
        # Pandas deprecated is_copy, and when we create new dataframes
        # from slices we do not want complaints. We always uses the
        # new frames knowing that they are separate from the original.
        self.pd_option_context = pd.option_context(
            "mode.chained_assignment", None
        )
        self.rc_context.__enter__()
        self.pd_option_context.__enter__()
        self._enter_ipython()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Exit matplotlib & pandas environments
        """
        import matplotlib.pyplot as plt

        if exc_type is None:
            if self.show:
                plt.show()
            else:
                plt.close(self.plot.figure)
        else:
            # There is an exception, close any figure
            if hasattr(self.plot, "figure"):
                plt.close(self.plot.figure)

        self.rc_context.__exit__(exc_type, exc_value, exc_traceback)
        self.pd_option_context.__exit__(exc_type, exc_value, exc_traceback)
        self._exit_ipython()
        delattr(self.plot.theme, "_targets")

    def _enter_ipython(self):
        """
        Setup ipython parameters in for the plot
        """
        ip = get_ipython()
        if not ip:
            return
        elif not hasattr(ip.config, "InlineBackend"):
            return

        for key, value in self._IPYTHON_CONFIG["InlineBackend"].items():
            if key not in ip.config.InlineBackend:  # pyright: ignore
                self._ip_config_inlinebackend[key] = key
                ip.run_line_magic("config", f"InlineBackend.{key} = {value!r}")

    def _exit_ipython(self):
        """
        Undo ipython parameters in for the plot
        """
        ip = get_ipython()
        if not ip:
            return
        elif not hasattr(ip.config, "InlineBackend"):
            return

        for key in self._ip_config_inlinebackend:
            del ip.config["InlineBackend"][key]  # pyright: ignore

        self._ip_config_inlinebackend = {}



import pandas as pd

from plotnine import (
    aes,
    element_text,
    facet_grid,
    geom_point,
    ggplot,
    labs,
    theme,
)
from plotnine.data import mtcars

c1 = "This is a sample caption"
c2 = """\
Lorem Ipsum is simply dummy text of the printing and typesetting industry.
Lorem Ipsum has been the industry's standard dummy text ever since the 1500s,
when an unknown printer took a galley of type and scrambled it to make
a type specimen book."""


df = pd.DataFrame({"x": [1, 2], "y": [3, 4], "cat": ["a", "b"]})


def test_labelling_with_colour():
    p = (
        ggplot(df, aes("x", "y", color="cat"))
        + geom_point()
        + labs(colour="Colour Title")
    )

    assert p == "labelling_with_colour"


def test_caption_simple():
    p = ggplot(mtcars, aes("wt", "mpg")) + geom_point() + labs(caption=c1)

    assert p == "caption_simple"


def test_caption_complex():
    p = (
        ggplot(mtcars, aes("wt", "mpg"))
        + geom_point()
        + labs(caption=c2)
        + facet_grid("am ~ vs")
        + theme(plot_caption=element_text(ha="left", size=12))
    )

    assert p == "caption_complex"



import warnings
from contextlib import suppress

import numpy as np
import pandas as pd

from ..exceptions import PlotnineError, PlotnineWarning
from ..utils import get_valid_kwargs


def predictdf(data, xseq, **params):
    """
    Make prediction on the data

    This is a general function responsible for dispatching
    to functions that do predictions for the specific models.
    """
    methods = {
        "lm": lm,
        "ols": lm,
        "wls": lm,
        "rlm": rlm,
        "glm": glm,
        "gls": gls,
        "lowess": lowess,
        "loess": loess,
        "mavg": mavg,
        "gpr": gpr,
    }

    method = params["method"]

    if isinstance(method, str):
        try:
            method = methods[method]
        except KeyError:
            msg = "Method should be one of {}"
            raise PlotnineError(msg.format(list(methods.keys())))

    if not callable(method):
        msg = (
            "'method' should either be a string or a function"
            "with the signature `func(data, xseq, **params)`"
        )
        raise PlotnineError()

    return method(data, xseq, **params)


def lm(data, xseq, **params):
    """
    Fit OLS / WLS if data has weight
    """
    import statsmodels.api as sm

    if params["formula"]:
        return lm_formula(data, xseq, **params)

    X = sm.add_constant(data["x"])
    Xseq = sm.add_constant(xseq)
    weights = data.get("weights", None)

    if weights is None:
        init_kwargs, fit_kwargs = separate_method_kwargs(
            params["method_args"], sm.OLS, sm.OLS.fit
        )
        model = sm.OLS(data["y"], X, **init_kwargs)
    else:
        if np.any(weights < 0):
            raise ValueError("All weights must be greater than zero.")
        init_kwargs, fit_kwargs = separate_method_kwargs(
            params["method_args"], sm.WLS, sm.WLS.fit
        )
        model = sm.WLS(data["y"], X, weights=data["weight"], **init_kwargs)

    results = model.fit(**fit_kwargs)
    data = pd.DataFrame({"x": xseq})
    data["y"] = results.predict(Xseq)

    if params["se"]:
        alpha = 1 - params["level"]
        prstd, iv_l, iv_u = wls_prediction_std(results, Xseq, alpha=alpha)
        data["se"] = prstd
        data["ymin"] = iv_l
        data["ymax"] = iv_u

    return data


def lm_formula(data, xseq, **params):
    """
    Fit OLS / WLS using a formula
    """
    import statsmodels.api as sm
    import statsmodels.formula.api as smf

    formula = params["formula"]
    eval_env = params["enviroment"]
    weights = data.get("weight", None)

    if weights is None:
        init_kwargs, fit_kwargs = separate_method_kwargs(
            params["method_args"], sm.OLS, sm.OLS.fit
        )
        model = smf.ols(formula, data, eval_env=eval_env, **init_kwargs)
    else:
        if np.any(weights < 0):
            raise ValueError("All weights must be greater than zero.")
        init_kwargs, fit_kwargs = separate_method_kwargs(
            params["method_args"], sm.OLS, sm.OLS.fit
        )
        model = smf.wls(
            formula, data, weights=weights, eval_env=eval_env, **init_kwargs
        )

    results = model.fit(**fit_kwargs)
    data = pd.DataFrame({"x": xseq})
    data["y"] = results.predict(data)

    if params["se"]:
        from patsy import dmatrices  # pyright: ignore

        _, predictors = dmatrices(formula, data, eval_env=eval_env)
        alpha = 1 - params["level"]
        prstd, iv_l, iv_u = wls_prediction_std(
            results, predictors, alpha=alpha
        )
        data["se"] = prstd
        data["ymin"] = iv_l
        data["ymax"] = iv_u
    return data


def rlm(data, xseq, **params):
    """
    Fit RLM
    """
    import statsmodels.api as sm

    if params["formula"]:
        return rlm_formula(data, xseq, **params)

    X = sm.add_constant(data["x"])
    Xseq = sm.add_constant(xseq)

    init_kwargs, fit_kwargs = separate_method_kwargs(
        params["method_args"], sm.RLM, sm.RLM.fit
    )
    model = sm.RLM(data["y"], X, **init_kwargs)
    results = model.fit(**fit_kwargs)

    data = pd.DataFrame({"x": xseq})
    data["y"] = results.predict(Xseq)

    if params["se"]:
        warnings.warn(
            "Confidence intervals are not yet implemented"
            " for RLM smoothing.",
            PlotnineWarning,
        )

    return data


def rlm_formula(data, xseq, **params):
    """
    Fit RLM using a formula
    """
    import statsmodels.api as sm
    import statsmodels.formula.api as smf

    eval_env = params["enviroment"]
    formula = params["formula"]
    init_kwargs, fit_kwargs = separate_method_kwargs(
        params["method_args"], sm.RLM, sm.RLM.fit
    )
    model = smf.rlm(formula, data, eval_env=eval_env, **init_kwargs)
    results = model.fit(**fit_kwargs)
    data = pd.DataFrame({"x": xseq})
    data["y"] = results.predict(data)

    if params["se"]:
        warnings.warn(
            "Confidence intervals are not yet implemented"
            " for RLM smoothing.",
            PlotnineWarning,
        )

    return data


def gls(data, xseq, **params):
    """
    Fit GLS
    """
    import statsmodels.api as sm

    if params["formula"]:
        return gls_formula(data, xseq, **params)

    X = sm.add_constant(data["x"])
    Xseq = sm.add_constant(xseq)

    init_kwargs, fit_kwargs = separate_method_kwargs(
        params["method_args"], sm.OLS, sm.OLS.fit
    )
    model = sm.GLS(data["y"], X, **init_kwargs)
    results = model.fit(**fit_kwargs)

    data = pd.DataFrame({"x": xseq})
    data["y"] = results.predict(Xseq)

    if params["se"]:
        alpha = 1 - params["level"]
        prstd, iv_l, iv_u = wls_prediction_std(results, Xseq, alpha=alpha)
        data["se"] = prstd
        data["ymin"] = iv_l
        data["ymax"] = iv_u

    return data


def gls_formula(data, xseq, **params):
    """
    Fit GLL using a formula
    """
    import statsmodels.api as sm
    import statsmodels.formula.api as smf

    eval_env = params["enviroment"]
    formula = params["formula"]
    init_kwargs, fit_kwargs = separate_method_kwargs(
        params["method_args"], sm.GLS, sm.GLS.fit
    )
    model = smf.gls(formula, data, eval_env=eval_env, **init_kwargs)
    results = model.fit(**fit_kwargs)
    data = pd.DataFrame({"x": xseq})
    data["y"] = results.predict(data)

    if params["se"]:
        from patsy import dmatrices  # pyright: ignore

        _, predictors = dmatrices(formula, data, eval_env=eval_env)
        alpha = 1 - params["level"]
        prstd, iv_l, iv_u = wls_prediction_std(
            results, predictors, alpha=alpha
        )
        data["se"] = prstd
        data["ymin"] = iv_l
        data["ymax"] = iv_u
    return data


def glm(data, xseq, **params):
    """
    Fit GLM
    """
    import statsmodels.api as sm

    if params["formula"]:
        return glm_formula(data, xseq, **params)

    X = sm.add_constant(data["x"])
    Xseq = sm.add_constant(xseq)

    init_kwargs, fit_kwargs = separate_method_kwargs(
        params["method_args"], sm.GLM, sm.GLM.fit
    )
    model = sm.GLM(data["y"], X, **init_kwargs)
    results = model.fit(**fit_kwargs)

    data = pd.DataFrame({"x": xseq})
    data["y"] = results.predict(Xseq)

    if params["se"]:
        prediction = results.get_prediction(Xseq)
        ci = prediction.conf_int(1 - params["level"])
        data["ymin"] = ci[:, 0]
        data["ymax"] = ci[:, 1]

    return data


def glm_formula(data, xseq, **params):
    """
    Fit with GLM formula
    """
    import statsmodels.api as sm
    import statsmodels.formula.api as smf

    eval_env = params["enviroment"]
    init_kwargs, fit_kwargs = separate_method_kwargs(
        params["method_args"], sm.GLM, sm.GLM.fit
    )
    model = smf.glm(params["formula"], data, eval_env=eval_env, **init_kwargs)
    results = model.fit(**fit_kwargs)
    data = pd.DataFrame({"x": xseq})
    data["y"] = results.predict(data)

    if params["se"]:
        df = pd.DataFrame({"x": xseq})
        prediction = results.get_prediction(df)
        ci = prediction.conf_int(1 - params["level"])
        data["ymin"] = ci[:, 0]
        data["ymax"] = ci[:, 1]
    return data


def lowess(data, xseq, **params):
    """
    Lowess fitting
    """
    import statsmodels.api as sm

    for k in ("is_sorted", "return_sorted"):
        with suppress(KeyError):
            del params["method_args"][k]
            warnings.warn(f"Smoothing method argument: {k}, has been ignored.")

    result = sm.nonparametric.lowess(
        data["y"],
        data["x"],
        frac=params["span"],
        is_sorted=True,
        **params["method_args"],
    )
    data = pd.DataFrame({"x": result[:, 0], "y": result[:, 1]})

    if params["se"]:
        warnings.warn(
            "Confidence intervals are not yet implemented"
            " for lowess smoothings.",
            PlotnineWarning,
        )

    return data


def loess(data, xseq, **params):
    """
    Loess smoothing
    """
    try:
        from skmisc.loess import loess as loess_klass
    except ImportError:
        raise PlotnineError("For loess smoothing, install 'scikit-misc'")

    try:
        weights = data["weight"]
    except KeyError:
        weights = None

    kwargs = params["method_args"]

    extrapolate = min(xseq) < min(data["x"]) or max(xseq) > max(data["x"])
    if "surface" not in kwargs and extrapolate:
        # Creates a loess model that allows extrapolation
        # when making predictions
        kwargs["surface"] = "direct"
        warnings.warn(
            "Making prediction outside the data range, "
            "setting loess control parameter `surface='direct'`.",
            PlotnineWarning,
        )

    if "span" not in kwargs:
        kwargs["span"] = params["span"]

    lo = loess_klass(data["x"], data["y"], weights, **kwargs)
    lo.fit()

    data = pd.DataFrame({"x": xseq})

    if params["se"]:
        alpha = 1 - params["level"]
        prediction = lo.predict(xseq, stderror=True)
        ci = prediction.confidence(alpha=alpha)
        data["se"] = prediction.stderr
        data["ymin"] = ci.lower
        data["ymax"] = ci.upper
    else:
        prediction = lo.predict(xseq, stderror=False)

    data["y"] = prediction.values

    return data


def mavg(data, xseq, **params):
    """
    Fit moving average
    """
    window = params["method_args"]["window"]

    # The first average comes after the full window size
    # has been swept over
    rolling = data["y"].rolling(**params["method_args"])
    y = rolling.mean()[window:]
    n = len(data)
    stderr = rolling.std()[window:]
    x = data["x"][window:]
    data = pd.DataFrame({"x": x, "y": y})
    data.reset_index(inplace=True, drop=True)

    if params["se"]:
        df = n - window  # Original - Used
        data["ymin"], data["ymax"] = tdist_ci(y, df, stderr, params["level"])
        data["se"] = stderr

    return data


def gpr(data, xseq, **params):
    """
    Fit gaussian process
    """
    try:
        from sklearn import gaussian_process
    except ImportError:
        raise PlotnineError(
            "To use gaussian process smoothing, "
            "You need to install scikit-learn."
        )

    kwargs = params["method_args"]
    if not kwargs:
        warnings.warn(
            "See sklearn.gaussian_process.GaussianProcessRegressor "
            "for parameters to pass in as 'method_args'",
            PlotnineWarning,
        )

    regressor = gaussian_process.GaussianProcessRegressor(**kwargs)
    X = np.atleast_2d(data["x"]).T
    n = len(data)
    Xseq = np.atleast_2d(xseq).T
    regressor.fit(X, data["y"])

    data = pd.DataFrame({"x": xseq})
    if params["se"]:
        y, stderr = regressor.predict(Xseq, return_std=True)
        data["y"] = y
        data["se"] = stderr
        data["ymin"], data["ymax"] = tdist_ci(
            y, n - 1, stderr, params["level"]
        )
    else:
        data["y"] = regressor.predict(Xseq, return_std=True)

    return data


def tdist_ci(x, df, stderr, level):
    """
    Confidence Intervals using the t-distribution
    """
    import scipy.stats as stats

    q = (1 + level) / 2
    if df is None:
        delta = stats.norm.ppf(q) * stderr
    else:
        delta = stats.t.ppf(q, df) * stderr
    return x - delta, x + delta


# Override wls_prediction_std from statsmodels to calculate the confidence
# interval instead of only the prediction interval
def wls_prediction_std(
    res, exog=None, weights=None, alpha=0.05, interval="confidence"
):
    """
    Calculate standard deviation and confidence interval

    Applies to WLS and OLS, not to general GLS,
    that is independently but not identically distributed observations

    Parameters
    ----------
    res : regression result instance
        results of WLS or OLS regression required attributes see notes
    exog : array_like (optional)
        exogenous variables for points to predict
    weights : scalar or array_like (optional)
        weights as defined for WLS (inverse of variance of observation)
    alpha : float (default: alpha = 0.05)
        confidence level for two-sided hypothesis
    interval : str
        Type of interval to compute. One of "confidence" or "prediction"

    Returns
    -------
    predstd : array_like, 1d
        standard error of prediction
        same length as rows of exog
    interval_l, interval_u : array_like
        lower und upper confidence bounds

    Notes
    -----
    The result instance needs to have at least the following
    res.model.predict() : predicted values or
    res.fittedvalues : values used in estimation
    res.cov_params() : covariance matrix of parameter estimates

    If exog is 1d, then it is interpreted as one observation,
    i.e. a row vector.

    testing status: not compared with other packages

    References
    ----------
    Greene p.111 for OLS, extended to WLS by analogy
    """
    import scipy.stats as stats

    # work around current bug:
    #    fit doesn't attach results to model, predict broken
    # res.model.results

    covb = res.cov_params()
    if exog is None:
        exog = res.model.exog
        predicted = res.fittedvalues
        if weights is None:
            weights = res.model.weights
    else:
        exog = np.atleast_2d(exog)
        if covb.shape[1] != exog.shape[1]:
            raise ValueError("wrong shape of exog")
        predicted = res.model.predict(res.params, exog)
        if weights is None:
            weights = 1.0
        else:
            weights = np.asarray(weights)
            if weights.size > 1 and len(weights) != exog.shape[0]:
                raise ValueError("weights and exog do not have matching shape")

    # full covariance:
    # predvar = res3.mse_resid + np.diag(np.dot(X2,np.dot(covb,X2.T)))
    # predication variance only
    predvar = res.mse_resid / weights
    ip = (exog * np.dot(covb, exog.T).T).sum(1)
    if interval == "confidence":
        predstd = np.sqrt(ip)
    elif interval == "prediction":
        predstd = np.sqrt(ip + predvar)
    else:
        raise ValueError(f"Unknown value for {interval=}")

    tppf = stats.t.isf(alpha / 2.0, res.df_resid)
    interval_u = predicted + tppf * predstd
    interval_l = predicted - tppf * predstd
    return predstd, interval_l, interval_u


def separate_method_kwargs(method_args, init_method, fit_method):
    """
    Categorise kwargs passed to the stat

    Some args are of the init method others for the fit method
    The separation is done by introspecting the init & fit methods
    """
    # inspect the methods
    init_kwargs = get_valid_kwargs(init_method, method_args)
    fit_kwargs = get_valid_kwargs(fit_method, method_args)

    # Warn about unknown kwargs
    known_kwargs = set(init_kwargs) | set(fit_kwargs)
    unknown_kwargs = set(method_args) - known_kwargs
    if unknown_kwargs:
        raise PlotnineError(
            "The following method arguments could not be recognised: "
            f"{list(unknown_kwargs)}"
        )
    return init_kwargs, fit_kwargs



from matplotlib.ticker import FixedFormatter, locale


class MyFixedFormatter(FixedFormatter):
    """
    Override MPL fixedformatter for better formatting
    """

    def format_data(self, value: float) -> str:
        """
        Return a formatted string representation of a number.
        """
        s = locale.format_string("%1.10e", (value,))
        return self.fix_minus(s)



import numpy as np
import pandas as pd

from plotnine import aes, geom_linerange, geom_pointrange, ggplot

n = 4
df = pd.DataFrame(
    {
        "x": range(n),
        "y": np.arange(n) + 0.5,
        "ymin": range(n),
        "ymax": range(1, n + 1),
        "z": range(n),
    }
)


def test_linerange_aesthetics():
    p = (
        ggplot(df, aes("x"))
        + geom_linerange(aes(ymin="ymin", ymax="ymax"), size=2)
        + geom_linerange(aes(ymin="ymin+1", ymax="ymax+1", alpha="z"), size=2)
        + geom_linerange(
            aes(ymin="ymin+2", ymax="ymax+2", linetype="factor(z)"), size=2
        )
        + geom_linerange(aes(ymin="ymin+3", ymax="ymax+3", color="z"), size=2)
        + geom_linerange(aes(ymin="ymin+4", ymax="ymax+4", size="z"))
    )
    assert p == "linerange_aesthetics"


def test_pointrange_aesthetics():
    p = (
        ggplot(df, aes("x"))
        + geom_pointrange(aes(y="y", ymin="ymin", ymax="ymax"), size=2)
        + geom_pointrange(
            aes(y="y+1", ymin="ymin+1", ymax="ymax+1", alpha="z"), size=2
        )
        + geom_pointrange(
            aes(y="y+2", ymin="ymin+2", ymax="ymax+2", linetype="factor(z)"),
            size=2,
        )
        + geom_pointrange(
            aes(y="y+3", ymin="ymin+3", ymax="ymax+3", color="z"), size=2
        )
        + geom_pointrange(aes(y="y+4", ymin="ymin+4", ymax="ymax+4", size="z"))
    )
    assert p == "pointrange_aesthetics"



from __future__ import annotations

import typing
from dataclasses import asdict, dataclass

from matplotlib.layout_engine import LayoutEngine

if typing.TYPE_CHECKING:
    from typing import Optional

    from matplotlib.backend_bases import RendererBase
    from matplotlib.offsetbox import AnchoredOffsetbox
    from matplotlib.text import Text

    from plotnine.typing import (
        Any,
        Axes,
        Facet,
        Figure,
        Ggplot,
        LegendPosition,
        Theme,
    )


@dataclass
class LayoutPack:
    """
    Objects required to compute the layout
    """

    axs: list[Axes]
    figure: Figure
    renderer: RendererBase
    theme: Theme
    facet: Facet
    axis_title_x: Optional[Text] = None
    axis_title_y: Optional[Text] = None
    legend: Optional[AnchoredOffsetbox] = None
    legend_position: Optional[LegendPosition] = None
    plot_caption: Optional[Text] = None
    plot_subtitle: Optional[Text] = None
    plot_title: Optional[Text] = None


class PlotnineLayoutEngine(LayoutEngine):
    """
    Implement geometry management for plotnine plots

    This layout manager automatically adjusts the location of
    objects placed around the plot panels and the subplot
    spacing parameters so that the plot fits cleanly within
    the figure area.
    """

    _adjust_compatible = True
    _colorbar_gridspec = False

    plot: Ggplot
    _theme_targets: dict[str, Any]

    def __init__(self, plot: Ggplot):
        self.plot = plot
        self._theme_targets = plot.theme._targets

    def execute(self, fig: Figure):
        from contextlib import nullcontext

        from ._plotnine_tight_layout import (
            get_plotnine_tight_layout,
            set_figure_artist_positions,
        )

        pack = self.pack_information()

        with getattr(pack.renderer, "_draw_disabled", nullcontext)():
            tparams = get_plotnine_tight_layout(pack)

        set_figure_artist_positions(pack, tparams)
        fig.subplots_adjust(**asdict(tparams.grid))

    def pack_information(self) -> LayoutPack:
        """
        Put together objects required to do the layout
        """
        _property = self.plot.theme.themeables.property

        def get(th: str) -> Any:
            """
            Return themeable target or None
            """
            return self._theme_targets.get(th, None)

        legend_position = _property("legend_position")
        if legend_position in ("none", "None"):
            legend_position = None

        return LayoutPack(
            axs=self.plot.axs,
            figure=self.plot.figure,
            renderer=self.plot.figure._get_renderer(),  # pyright: ignore
            theme=self.plot.theme,
            facet=self.plot.facet,
            axis_title_x=get("axis_title_x"),
            axis_title_y=get("axis_title_y"),
            legend=get("legend_background"),
            legend_position=legend_position,
            plot_caption=get("plot_caption"),
            plot_subtitle=get("plot_subtitle"),
            plot_title=get("plot_title"),
        )



import numpy as np
import pandas as pd

from ..doctools import document
from ..mapping.evaluation import after_stat
from .stat import stat


@document
class stat_ecdf(stat):
    """
    Emperical Cumulative Density Function

    {usage}

    Parameters
    ----------
    {common_parameters}
    n  : int (default: None)
        This is the number of points to interpolate with.
        If :py:`None`, do not interpolate.

    See Also
    --------
    plotnine.geoms.geom_step
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'x'  # x in the data
        'y'  # cumulative density corresponding to x

    """

    REQUIRED_AES = {"x"}
    DEFAULT_PARAMS = {
        "geom": "step",
        "position": "identity",
        "na_rm": False,
        "n": None,
    }
    DEFAULT_AES = {"y": after_stat("y")}
    CREATES = {"y"}

    @classmethod
    def compute_group(cls, data, scales, **params):
        from statsmodels.distributions.empirical_distribution import ECDF

        # If n is None, use raw values; otherwise interpolate
        if params["n"] is None:
            x = np.unique(data["x"])
        else:
            x = np.linspace(data["x"].min(), data["x"].max(), params["n"])

        y = ECDF(data["x"])(x)
        res = pd.DataFrame({"x": x, "y": y})
        return res



from __future__ import annotations

import typing
from contextlib import suppress
from typing import List

from ..iapi import strip_draw_info, strip_label_details

if typing.TYPE_CHECKING:
    from plotnine.iapi import layout_details
    from plotnine.typing import (
        Axes,
        Facet,
        Layout,
        StripPosition,
        Theme,
    )


class strip:
    """
    A strip

    This class exists to have in one place all that is required to draw
    strip text onto an axes. As Matplotlib does not have a layout manager
    that makes it easy to adorn an axes with artists, we have to compute
    the space required for the text and the background strip on which it
    is drawn. This is very finicky and fails once the facets become
    complicated.
    """

    position: StripPosition
    label_info: strip_label_details

    def __init__(
        self,
        vars: list[str],
        layout_info: layout_details,
        facet: Facet,
        ax: Axes,
        position: StripPosition,
    ):
        self.vars = vars
        self.ax = ax
        self.position = position
        self.facet = facet
        self.figure = facet.figure
        self.theme = facet.theme
        label_info = strip_label_details.make(layout_info, vars, position)
        self.label_info = facet.labeller(label_info)
        self.draw_info = self.get_draw_info()

    def get_draw_info(self) -> strip_draw_info:
        """
        Get information required to draw strips

        Returns
        -------
        out :
            A structure with all the coordinates (x, y) required
            to draw the strip text and the background box
            (box_x, box_y, box_width, box_height).
        """
        _property = self.theme.themeables.property
        position = self.position
        if position == "top":
            # The x & y values are just starting locations
            # The final location is determined by the layout manager.
            x, y = 0, 1
            ha, va = "center", "bottom"
            rotation = _property("strip_text_x", "rotation")
            box_width = 1
            box_height = 0  # Determined by the text size
            # TODO: Allow two unique paddings for either side.
            # Requires implementing an mpl.patches.boxstyle that recognises
            # two padding values.
            strip_text_margin = _property("strip_text_x", "margin").get_as(
                "b", "lines"
            )
            strip_align = _property("strip_align_x")

            # x & width properties of the background slide and
            # shrink the strip horizontally.
            with suppress(KeyError):
                x = _property("strip_text_x", "x")
            with suppress(KeyError):
                box_width = _property("strip_background_x", "width")

        elif position == "right":
            # The x & y values are just starting locations
            # The final location is determined by the layout manager.
            x, y = 1, 0
            ha, va = "left", "center"
            rotation = _property("strip_text_y", "rotation")
            box_width = 0  # Determine by the text height
            box_height = 1
            # TODO: Allow two unique paddings for either side.
            # Requires implementing an mpl.patches.boxstyle that recognises
            # two padding values.
            strip_text_margin = _property("strip_text_y", "margin").get_as(
                "r", "lines"
            )
            strip_align = _property("strip_align_y")

            # y & height properties of the background slide and
            # shrink the strip vertically.
            with suppress(KeyError):
                y = _property("strip_text_y", "y")
            with suppress(KeyError):
                box_height = _property("strip_background_y", "height")
        else:
            raise ValueError(f"Unknown position for strip text: {position!r}")

        return strip_draw_info(
            x=x,
            y=y,
            ha=ha,
            va=va,
            box_width=box_width,
            box_height=box_height,
            strip_text_margin=strip_text_margin,
            strip_align=strip_align,
            position=position,
            label=self.label_info.text(),
            ax=self.ax,
            rotation=rotation,
        )

    def draw(self):
        """
        Create a background patch and put a label on it
        """

        from .._mpl.text import SText

        _targets = self.theme._targets
        text = SText(self.draw_info)
        rect = text.spatch

        self.ax.add_artist(text)
        self.ax.add_artist(rect)

        for key in (
            "strip_text_x",
            "strip_text_y",
            "strip_background_x",
            "strip_background_y",
        ):
            if key not in _targets:
                _targets[key] = []

        if self.draw_info.position == "right":
            _targets["strip_background_y"].append(rect)
            _targets["strip_text_y"].append(text)
        else:
            _targets["strip_background_x"].append(rect)
            _targets["strip_text_x"].append(text)


class Strips(List[strip]):
    """
    List of strips for a plot
    """

    facet: Facet

    @staticmethod
    def from_facet(facet: Facet) -> Strips:
        new = Strips()
        new.facet = facet
        return new

    @property
    def axs(self) -> list[Axes]:
        return self.facet.axs

    @property
    def layout(self) -> Layout:
        return self.facet.layout

    @property
    def theme(self) -> Theme:
        return self.facet.theme

    @property
    def top_strips(self) -> Strips:
        return Strips([s for s in self if s.position == "top"])

    @property
    def right_strips(self) -> Strips:
        return Strips([s for s in self if s.position == "right"])

    def draw(self):
        for s in self:
            s.draw()

    def generate(self):
        """
        Calculate the box information for all strips

        It is stored in self.strip_info
        """
        for layout_info in self.layout.get_details():
            ax = self.axs[layout_info.panel_index]
            lst = self.facet.make_ax_strips(layout_info, ax)
            self.extend(lst)



from __future__ import annotations

import typing

from ..doctools import document
from ..utils import alias
from .scale import scale_continuous, scale_discrete

if typing.TYPE_CHECKING:
    from typing import Any, Literal, Sequence


class MapTrainMixin:
    """
    Override map and train methods
    """

    guide: Literal["legend"] | None = None

    def map(self, x: Sequence[Any]) -> Sequence[Any]:
        """
        Identity map

        Notes
        -----
        Identity scales bypass the palette completely since the
        map is the identity function.
        """
        return x

    def train(self, x):
        # do nothing if no guide,
        # otherwise train so we know what breaks to use
        if self.guide is None:
            return

        return super().train(x)  # pyright: ignore


@document
class scale_color_identity(MapTrainMixin, scale_discrete):
    """
    No color scaling

    Parameters
    ----------
    {superclass_parameters}
    guide : None | 'legend'
        Whether to include a legend. Default is None.
    """

    _aesthetics = ["color"]


@document
class scale_fill_identity(scale_color_identity):
    """
    No color scaling

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


@document
class scale_shape_identity(MapTrainMixin, scale_discrete):
    """
    No shape scaling

    Parameters
    ----------
    {superclass_parameters}
    guide : None | 'legend'
        Whether to include a legend. Default is None.
    """

    _aesthetics = ["shape"]


@document
class scale_linetype_identity(MapTrainMixin, scale_discrete):
    """
    No linetype scaling

    Parameters
    ----------
    {superclass_parameters}
    guide : None | 'legend'
        Whether to include a legend. Default is None.
    """

    _aesthetics = ["linetype"]


@document
class scale_alpha_identity(MapTrainMixin, scale_continuous):
    """
    No alpha scaling

    Parameters
    ----------
    {superclass_parameters}
    guide : None | 'legend'
        Whether to include a legend. Default is None.
    """

    _aesthetics = ["alpha"]


@document
class scale_size_identity(MapTrainMixin, scale_continuous):
    """
    No size scaling

    Parameters
    ----------
    {superclass_parameters}
    guide : None | 'legend'
        Whether to include a legend. Default is None.
    """

    _aesthetics = ["size"]


# American to British spelling
alias("scale_colour_identity", scale_color_identity)



from __future__ import annotations

import numbers
import typing

import numpy as np
import pandas as pd
import pandas.api.types as pdtypes

from ..exceptions import PlotnineError

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.typing import EvalEnvironment

    from . import aes


__all__ = ["after_stat", "after_scale", "stage"]

_TPL_EVAL_FAIL = """\
Could not evaluate the '{}' mapping: '{}' \
(original error: {})"""

_TPL_BAD_EVAL_TYPE = """\
The '{}' mapping: '{}' produced a value of type '{}',\
but only single items and lists/arrays can be used. \
(original error: {})"""


class stage:
    """
    Stage allows you evaluating mapping at more than one stage

    You can evaluate an expression of a variable in a dataframe, and
    later evaluate an expression that modifies the values mapped to
    the scale.

    Parameters
    ----------
    start : expression | array_like | scalar
        Aesthetic expression using primary variables from the layer
        data.
    after_stat : expression
        Aesthetic expression using variables calculated by the stat.
    after_scale : expression
        Aesthetic expression using aesthetics of the layer.
    """

    def __init__(self, start=None, after_stat=None, after_scale=None):
        self.start = start
        self.after_stat = after_stat
        self.after_scale = after_scale

    def __repr__(self):
        """
        Repr for staged mapping
        """
        # Shorter representation when the mapping happens at a
        # single stage
        if self.after_stat is None and self.after_scale is None:
            return f"{repr(self.start)}"
        if self.start is None and self.after_scale is None:
            return f"after_stat({repr(self.after_stat)})"
        if self.start is None and self.after_stat is None:
            return f"after_scale({repr(self.after_scale)})"
        return (
            f"stage(start={repr(self.start)}, "
            f"after_stat={repr(self.after_stat)}, "
            f"after_scale={repr(self.after_scale)})"
        )


def after_stat(x):
    """
    Evaluate mapping after statistic has been calculated

    Parameters
    ----------
    x : str
        An expression

    See Also
    --------
    :func:`after_scale` : For how to alter aesthetics after the data has been
        mapped by the scale.
    :class:`stage` : For how to map to aesthetics at more than one stage of
        the plot building pipeline.
    """
    return stage(after_stat=x)


def after_scale(x):
    """
    Evaluate mapping after variable has been mapped to the scale

    This gives the user a chance to alter the value of a variable
    in the final units of the scale e.g. the rgb hex color.

    Parameters
    ----------
    x : str
        An expression

    See Also
    --------
    :func:`after_stat` : For how to map aesthetics to variable calculated
        by the stat
    :class:`stage` : For how to map to aesthetics at more than one stage of
        the plot building pipeline.
    """
    return stage(after_scale=x)


def reorder(x, y, fun=np.median, ascending=True):
    """
    Reorder categorical by sorting along another variable

    It is the order of the categories that changes. Values in x
    are grouped by categories and summarised to determine the
    new order.

    Credit: Copied from plydata

    Parameters
    ----------
    x : list-like
        Values that will make up the categorical.
    y : list-like
        Values by which ``c`` will be ordered.
    fun : callable
        Summarising function to ``x`` for each category in ``c``.
        Default is the *median*.
    ascending : bool
        If ``True``, the ``c`` is ordered in ascending order of ``x``.

    Examples
    --------
    >>> c = list('abbccc')
    >>> x = [11, 2, 2, 3, 33, 3]
    >>> cat_reorder(c, x)
    [a, b, b, c, c, c]
    Categories (3, object): [b, c, a]
    >>> cat_reorder(c, x, fun=max)
    [a, b, b, c, c, c]
    Categories (3, object): [b, a, c]
    >>> cat_reorder(c, x, fun=max, ascending=False)
    [a, b, b, c, c, c]
    Categories (3, object): [c, a, b]
    >>> c_ordered = pd.Categorical(c, ordered=True)
    >>> cat_reorder(c_ordered, x)
    [a, b, b, c, c, c]
    Categories (3, object): [b < c < a]
    >>> cat_reorder(c + ['d'], x)
    Traceback (most recent call last):
        ...
    ValueError: Lengths are not equal. len(c) is 7 and len(x) is 6.
    """
    if len(x) != len(y):
        raise ValueError(f"Lengths are not equal. {len(x)=}, {len(x)=}")
    summary = (
        pd.Series(y).groupby(x).apply(fun).sort_values(ascending=ascending)
    )
    cats = summary.index.to_list()
    return pd.Categorical(x, categories=cats)


# These are function that can be called by the user inside the aes()
# mapping. This is meant to make the variable transformations as easy
# as they are in ggplot2
AES_INNER_NAMESPACE = {"factor": pd.Categorical, "reorder": reorder}


def evaluate(
    aesthetics: aes | dict[str, Any], data: pd.DataFrame, env: EvalEnvironment
) -> pd.DataFrame:
    """
    Evaluate aesthetics

    Parameters
    ----------
    aesthetics : dict-like
        Aesthetics to evaluate. They must be of the form {name: expr}
    data : pd.DataFrame
        Dataframe whose columns are/may-be variables in the aesthetic
        expressions i.e. it is a namespace with variables.
    env : ~patsy.Eval.Environment
        Environment in which the aesthetics are evaluated

    Returns
    -------
    evaled : pd.DataFrame
        Dataframe of the form {name: result}, where each column is the
        result from evaluating an expression.

    Examples
    --------
    >>> import patsy
    >>> var1 = 2
    >>> env = patsy.eval.EvalEnvironment.capture()
    >>> df = pd.DataFrame({'x': range(1, 6)})
    >>> aesthetics = {'y': 'x**var1'}
    >>> evaluate(aesthetics, df, env)
        y
    0   1
    1   4
    2   9
    3  16
    4  25
    """
    env = env.with_outer_namespace(AES_INNER_NAMESPACE)

    # Store evaluation results in a dict column in a dict
    evaled = {}

    # If a column name is not in the data, it is evaluated/transformed
    # in the environment of the call to ggplot
    for ae, col in aesthetics.items():
        if isinstance(col, str):
            if col in data:
                evaled[ae] = data[col]
            else:
                try:
                    new_val = env.eval(col, inner_namespace=data)
                except Exception as e:
                    raise PlotnineError(_TPL_EVAL_FAIL.format(ae, col, str(e)))

                try:
                    evaled[ae] = new_val
                except Exception as e:
                    raise PlotnineError(
                        _TPL_BAD_EVAL_TYPE.format(
                            ae, col, str(type(new_val)), str(e)
                        )
                    )
        elif pdtypes.is_list_like(col):
            n = len(col)
            if len(data) and n != len(data) and n != 1:
                raise PlotnineError(
                    "Aesthetics must either be length one, "
                    "or the same length as the data"
                )
            evaled[ae] = col
        elif is_known_scalar(col):
            if not len(evaled):
                col = [col]
            evaled[ae] = col
        else:
            msg = "Do not know how to deal with aesthetic '{}'"
            raise PlotnineError(msg.format(ae))

    # Using `type` preserves the subclass of pd.DataFrame
    index = data.index if len(data.index) else None
    evaled = type(data)(data=evaled, index=index)
    return evaled


def is_known_scalar(value):
    """
    Return True if value is a type we expect in a dataframe
    """

    def _is_datetime_or_timedelta(value):
        # Using pandas.Series helps catch python, numpy and pandas
        # versions of these types
        return pd.Series(value).dtype.kind in ("M", "m")

    return not np.iterable(value) and (
        isinstance(value, numbers.Number) or _is_datetime_or_timedelta(value)
    )



from __future__ import annotations

import typing

import numpy as np

from ..coords import coord_flip
from ..doctools import document
from ..utils import SIZE_FACTOR, make_line_segments, to_rgba
from .geom import geom
from .geom_path import geom_path

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord


@document
class geom_rug(geom):
    """
    Marginal rug plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    sides : str (default: bl)
        Sides onto which to draw the marks. Any combination
        chosen from the characters ``btlr``, for *bottom*, *top*,
        *left* or *right* side marks.
    length: float
        length of marks in fractions of
        horizontal/vertical panel size (default 0.03)
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "black",
        "size": 0.5,
        "linetype": "solid",
    }
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "sides": "bl",
        "length": 0.03,
    }
    draw_legend = staticmethod(geom_path.draw_legend)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        from matplotlib.collections import LineCollection

        data = coord.transform(data, panel_params)
        sides = params["sides"]

        # coord_flip does not flip the side(s) on which the rugs
        # are plotted. We do the fliping here
        if isinstance(coord, coord_flip):
            t = str.maketrans("tblr", "rlbt")
            sides = sides.translate(t)

        data["size"] *= SIZE_FACTOR

        has_x = "x" in data.columns
        has_y = "y" in data.columns

        if has_x or has_y:
            n = len(data)
        else:
            return

        rugs = []
        xmin, xmax = panel_params.x.range
        ymin, ymax = panel_params.y.range
        xheight = (xmax - xmin) * params["length"]
        yheight = (ymax - ymin) * params["length"]

        if has_x:
            if "b" in sides:
                x = np.repeat(data["x"].to_numpy(), 2)
                y = np.tile([ymin, ymin + yheight], n)
                rugs.extend(make_line_segments(x, y, ispath=False))

            if "t" in sides:
                x = np.repeat(data["x"].to_numpy(), 2)
                y = np.tile([ymax - yheight, ymax], n)
                rugs.extend(make_line_segments(x, y, ispath=False))

        if has_y:
            if "l" in sides:
                x = np.tile([xmin, xmin + xheight], n)
                y = np.repeat(data["y"].to_numpy(), 2)
                rugs.extend(make_line_segments(x, y, ispath=False))

            if "r" in sides:
                x = np.tile([xmax - xheight, xmax], n)
                y = np.repeat(data["y"].to_numpy(), 2)
                rugs.extend(make_line_segments(x, y, ispath=False))

        color = to_rgba(data["color"], data["alpha"])
        coll = LineCollection(
            rugs,
            edgecolor=color,
            linewidth=data["size"],
            linestyle=data["linetype"],
            zorder=params["zorder"],
            rasterized=params["raster"],
        )
        ax.add_collection(coll)



from __future__ import annotations

import hashlib
import types
import typing
from contextlib import suppress
from itertools import islice
from warnings import warn

import numpy as np
import pandas as pd

from ..exceptions import PlotnineError, PlotnineWarning
from ..geoms import geom_text
from ..mapping.aes import rename_aesthetics
from ..utils import SIZE_FACTOR, remove_missing
from .guide import guide

if typing.TYPE_CHECKING:
    from typing import Optional

    from plotnine.typing import TupleInt2


# See guides.py for terminology


class guide_legend(guide):
    """
    Legend guide

    Parameters
    ----------
    nrow : int
        Number of rows of legends.
    ncol : int
        Number of columns of legends.
    byrow : bool
        Whether to fill the legend row-wise or column-wise.
    keywidth : float
        Width of the legend key.
    keyheight : float
        Height of the legend key.
    kwargs : dict
        Parameters passed on to :class:`.guide`
    """

    # general
    nrow: int = -1
    ncol: int = -1
    byrow = False

    # key
    keywidth: Optional[int] = None
    keyheight: Optional[int] = None

    # parameter
    available_aes = {"any"}

    def train(self, scale, aesthetic=None):
        """
        Create the key for the guide

        The key is a dataframe with two columns:
            - scale name : values
            - label : labels for each value

        scale name is one of the aesthetics
        ['x', 'y', 'color', 'fill', 'size', 'shape', 'alpha',
         'stroke']

        Returns this guide if trainning is successful and None
        if it fails
        """
        if aesthetic is None:
            aesthetic = scale.aesthetics[0]

        breaks = scale.get_bounded_breaks()
        if not breaks:
            return None

        key = pd.DataFrame(
            {aesthetic: scale.map(breaks), "label": scale.get_labels(breaks)}
        )

        if len(key) == 0:
            return None

        self.key = key

        # create a hash of the important information in the guide
        labels = " ".join(str(x) for x in self.key["label"])
        info = "\n".join(
            [self.title, labels, str(self.direction), self.__class__.__name__]
        )
        self.hash = hashlib.md5(info.encode("utf-8")).hexdigest()
        return self

    def merge(self, other):
        """
        Merge overlapped guides

        For example::

            from ggplot import *
            gg = ggplot(aes(x='cut', fill='cut', color='cut'), data=diamonds)
            gg + stat_bin()

        This would create similar guides for fill and color where only
        a single guide would do
        """
        self.key = pd.merge(self.key, other.key)
        duplicated = set(self.override_aes) & set(other.override_aes)
        if duplicated:
            warn("Duplicated override_aes is ignored.", PlotnineWarning)
        self.override_aes.update(other.override_aes)
        for ae in duplicated:
            del self.override_aes[ae]
        return self

    def create_geoms(self, plot):
        """
        Make information needed to draw a legend for each of the layers.

        For each layer, that information is a dictionary with the geom
        to draw the guide together with the data and the parameters that
        will be used in the call to geom.
        """
        # A layer either contributes to the guide, or it does not. The
        # guide entries may be ploted in the layers
        self.glayers = []
        for l in plot.layers:
            exclude = set()
            if isinstance(l.show_legend, dict):
                l.show_legend = rename_aesthetics(l.show_legend)
                exclude = {ae for ae, val in l.show_legend.items() if not val}
            elif l.show_legend not in (None, True):
                continue

            matched = self.legend_aesthetics(l, plot)

            # This layer does not contribute to the legend
            if not set(matched) - exclude:
                continue

            data = self.key[matched].copy()

            # Modify aesthetics
            try:
                data = l.use_defaults(data)
            except PlotnineError:
                warn(
                    "Failed to apply `after_scale` modifications "
                    "to the legend.",
                    PlotnineWarning,
                )
                data = l.use_defaults(data, aes_modifiers={})

            # override.aes in guide_legend manually changes the geom
            for ae in set(self.override_aes) & set(data.columns):
                data[ae] = self.override_aes[ae]

            data = remove_missing(
                data,
                l.geom.params["na_rm"],
                list(l.geom.REQUIRED_AES | l.geom.NON_MISSING_AES),
                f"{l.geom.__class__.__name__} legend",
            )
            self.glayers.append(
                types.SimpleNamespace(geom=l.geom, data=data, layer=l)
            )
        if not self.glayers:
            return None
        return self

    def _calculate_rows_and_cols(self) -> TupleInt2:
        nrow, ncol = -1, -1
        nbreak = len(self.key)

        if hasattr(self, "nrow"):
            nrow = self.nrow

        if hasattr(self, "ncol"):
            ncol = self.ncol

        if nrow != -1 and ncol != -1:
            if nrow * ncol < nbreak:
                raise PlotnineError(
                    "nrow x ncol need to be larger "
                    "than the number of breaks"
                )
            return nrow, ncol

        if nrow == -1 and ncol == -1:
            if self.direction == "horizontal":
                nrow = int(np.ceil(nbreak / 5))
            else:
                ncol = int(np.ceil(nbreak / 20))

        if nrow == -1:
            nrow = int(np.ceil(nbreak / ncol))
        elif ncol == -1:
            ncol = int(np.ceil(nbreak / nrow))

        return nrow, ncol

    def _set_defaults(self, theme):
        guide._set_defaults(self, theme)
        _property = theme.themeables.property
        self.nrow, self.ncol = self._calculate_rows_and_cols()
        nbreak = len(self.key)

        # key width and key height for each legend entry
        #
        # Take a peak into data['size'] to make sure the
        # legend dimensions are big enough
        """
        >>> gg = ggplot(diamonds, aes(x='cut', y='clarity'))
        >>> gg = gg + stat_sum(aes(group='cut'))
        >>> gg + scale_size(range=(3, 25))

        Note the different height sizes for the entries
        """

        # FIXME: This should be in the geom instead of having
        # special case conditional branches
        def determine_side_length(initial_size):
            default_pad = initial_size * 0.5
            # default_pad = 0
            size = np.ones(nbreak) * initial_size
            for i in range(nbreak):
                for gl in self.glayers:
                    _size = 0
                    pad = default_pad
                    # Full size of object to appear in the
                    # legend key
                    with suppress(IndexError):
                        if "size" in gl.data:
                            _size = gl.data["size"].iloc[i] * SIZE_FACTOR
                            if "stroke" in gl.data:
                                _size += (
                                    2 * gl.data["stroke"].iloc[i] * SIZE_FACTOR
                                )

                        # special case, color does not apply to
                        # border/linewidth
                        if isinstance(gl.geom, geom_text):
                            pad = 0
                            if _size < initial_size:
                                continue

                        try:
                            # color(edgecolor) affects size(linewidth)
                            # When the edge is not visible, we should
                            # not expand the size of the keys
                            if gl.data["color"].iloc[i] is not None:
                                size[i] = np.max([_size + pad, size[i]])
                        except KeyError:
                            break

            return size

        # keysize
        if self.keywidth is None:
            width = determine_side_length(_property("legend_key_width"))
            if self.direction == "vertical":
                width[:] = width.max()
            self._keywidth = width
        else:
            self._keywidth = [self.keywidth] * nbreak

        if self.keyheight is None:
            height = determine_side_length(_property("legend_key_height"))
            if self.direction == "horizontal":
                height[:] = height.max()
            self._keyheight = height
        else:
            self._keyheight = [self.keyheight] * nbreak

    def draw(self):
        """
        Draw guide

        Returns
        -------
        out : matplotlib.offsetbox.Offsetbox
            A drawing of this legend
        """
        from matplotlib.offsetbox import HPacker, TextArea, VPacker

        from .._mpl.offsetbox import ColoredDrawingArea

        obverse = slice(0, None)
        reverse = slice(None, None, -1)
        nbreak = len(self.key)
        _targets = self.theme._targets

        # When there is more than one guide, we keep
        # record of all of them using lists
        if "legend_title" not in _targets:
            _targets["legend_title"] = []
        if "legend_text_legend" not in _targets:
            _targets["legend_key"] = []
            _targets["legend_text_legend"] = []

        # title
        title_box = TextArea(self.title, textprops={"color": "black"})
        _targets["legend_title"].append(title_box)

        # labels
        labels = []
        for item in self.key["label"]:
            if isinstance(item, float) and float.is_integer(item):
                item = int(item)  # 1.0 to 1
            va = "center" if self.label_position == "top" else "baseline"
            ta = TextArea(item, textprops={"color": "black", "va": va})
            labels.append(ta)
            _targets["legend_text_legend"].extend(labels)

        # Drawings
        drawings = []
        for i in range(nbreak):
            da = ColoredDrawingArea(
                self._keywidth[i], self._keyheight[i], 0, 0, color="white"
            )
            # overlay geoms
            for gl in self.glayers:
                with suppress(IndexError):
                    data = gl.data.iloc[i]
                    da = gl.geom.draw_legend(data, da, gl.layer)
            drawings.append(da)
        _targets["legend_key"].append(drawings)

        # Match Drawings with labels to create the entries
        lookup = {
            "right": (HPacker, reverse),
            "left": (HPacker, obverse),
            "bottom": (VPacker, reverse),
            "top": (VPacker, obverse),
        }
        packer, slc = lookup[self.label_position]
        entries = []
        for d, l in zip(drawings, labels):
            e = packer(
                children=[l, d][slc],
                sep=self._label_margin,
                align="center",
                pad=0,
            )
            entries.append(e)

        # Put the entries together in rows or columns
        # A chunk is either a row or a column of entries
        # for a single legend
        if self.byrow:
            chunk_size, packers = self.ncol, [HPacker, VPacker]
            sep1 = self._legend_entry_spacing_x
            sep2 = self._legend_entry_spacing_y
        else:
            chunk_size, packers = self.nrow, [VPacker, HPacker]
            sep1 = self._legend_entry_spacing_y
            sep2 = self._legend_entry_spacing_x

        if self.reverse:
            entries = entries[::-1]
        chunks = []
        for i in range(len(entries)):
            start = i * chunk_size
            stop = start + chunk_size
            s = islice(entries, start, stop)
            chunks.append(list(s))
            if stop >= len(entries):
                break

        chunk_boxes = []
        for chunk in chunks:
            d1 = packers[0](children=chunk, align="left", sep=sep1, pad=0)
            chunk_boxes.append(d1)

        # Put all the entries (row & columns) together
        entries_box = packers[1](
            children=chunk_boxes, align="baseline", sep=sep2, pad=0
        )

        # Put the title and entries together
        packer, slc = lookup[self.title_position]
        children = [title_box, entries_box][slc]
        box = packer(
            children=children,
            sep=self._title_margin,
            align=self._title_align,
            pad=self._legend_margin,
        )

        return box
